import warnings
from abc import ABC
from abc import abstractmethod
from itertools import product
from typing import Optional
from typing import Sequence
from typing import TypeVar
from typing import Union

import numpy as np
from scipy.integrate import odeint
from scipy.integrate import solve_ivp
from scipy.interpolate import interp1d
from sklearn.base import BaseEstimator
from sklearn.metrics import r2_score
from sklearn.utils.validation import check_is_fitted
from typing_extensions import Self

from .differentiation import BaseDifferentiation
from .differentiation import FiniteDifference
from .feature_library import PolynomialLibrary
from .feature_library.base import BaseFeatureLibrary

try:  # Waiting on PEP 690 to lazy import CVXPY
    from .optimizers import SINDyPI

    sindy_pi_flag = True
except ImportError:
    sindy_pi_flag = False
from .optimizers import STLSQ
from .optimizers.base import _BaseOptimizer
from .optimizers.base import BaseOptimizer
from .utils import AxesArray
from .utils import comprehend_axes
from .utils import concat_sample_axis
from .utils import drop_nan_samples
from .utils import SampleConcatter
from .utils import validate_control_variables
from .utils import validate_no_reshape


TrajectoryType = TypeVar("TrajectoryType", list[np.ndarray], np.ndarray)


class _BaseSINDy(BaseEstimator, ABC):

    feature_library: BaseFeatureLibrary
    optimizer: _BaseOptimizer
    # Hacks to remove later
    feature_names: Optional[list[str]]
    n_control_features_: int = 0

    @abstractmethod
    def fit(self, x: TrajectoryType, t: TrajectoryType, *args, **kwargs) -> Self:
        ...

    @abstractmethod
    def simulate(self, x0: np.ndarray, t: np.ndarray) -> np.ndarray:
        ...

    @abstractmethod
    def score(
        self, x: TrajectoryType, t: TrajectoryType, x_dot: TrajectoryType
    ) -> float:
        ...

    def _fit_shape(self):
        """Assign shape attributes for the system that are used post-fit"""
        self.n_features_in_ = self.feature_library.n_features_in_
        self.n_output_features_ = self.feature_library.n_output_features_
        if self.feature_names is None:
            feature_names = []
            for i in range(self.n_features_in_ - self.n_control_features_):
                feature_names.append("x" + str(i))
            for i in range(self.n_control_features_):
                feature_names.append("u" + str(i))
            self.feature_names = feature_names

    def predict(self, x, u=None):
        """
        Predict the right hand side of the dynamical system

        Parameters
        ----------
        x: array-like or list of array-like, shape (n_samples, n_input_features)
            Samples.

        u: array-like or list of array-like, shape(n_samples, n_control_features), \
                (default None)
            Control variables. If ``multiple_trajectories==True`` then u
            must be a list of control variable data from each trajectory. If the
            model was fit with control variables then u is not optional.

        Returns
        -------
        result: array-like or list of array-like, shape (n_samples, n_input_features)
            Predicted right hand side of the dynamical system
        """
        if not _check_multiple_trajectories(x, None, u):
            x, _, _, u = _adapt_to_multiple_trajectories(x, None, None, u)
            multiple_trajectories = False
        else:
            multiple_trajectories = True

        x, _, u = _comprehend_and_validate_inputs(x, 1, None, u, self.feature_library)

        check_is_fitted(self)
        if self.n_control_features_ > 0 and u is None:
            raise TypeError("Model was fit using control variables, so u is required")
        if self.n_control_features_ == 0 and u is not None:
            warnings.warn(
                "Control variables u were ignored because control variables were"
                " not used when the model was fit"
            )
            u = None
        if u is not None:
            u = validate_control_variables(x, u)
            x = [np.concatenate((xi, ui), axis=xi.ax_coord) for xi, ui in zip(x, u)]

        x_feat = self.feature_library.transform(x)
        x_feat = [SampleConcatter().fit_transform([xi]) for xi in x_feat]

        result = [self.optimizer.predict(xi) for xi in x_feat]
        result = [
            self.feature_library.reshape_samples_to_spatial_grid(pred)
            for pred in result
        ]

        # Kept for backwards compatibility.
        if not multiple_trajectories:
            return result[0]
        return result

    def coefficients(self):
        """
        Get an array of the coefficients learned by SINDy model.

        Returns
        -------
        coef: np.ndarray, shape (n_input_features, n_output_features)
            Learned coefficients of the SINDy model.
            Equivalent to :math:`\\Xi^\\top` in the literature.
        """
        check_is_fitted(self)
        return self.optimizer.coef_

    def equations(self, precision: int = 3) -> list[str]:
        """
        Get the right hand sides of the SINDy model equations.

        Parameters
        ----------
        precision: int, optional (default 3)
            Number of decimal points to include for each coefficient in the
            equation.

        Returns
        -------
        equations: list of strings
            List of strings representing the SINDy model equations for each
            input feature.
        """
        check_is_fitted(self)
        sys_coord_names = self.feature_names
        feat_names = self.feature_library.get_feature_names(sys_coord_names)

        def term(c, name):
            rounded_coef = np.round(c, precision)
            if rounded_coef == 0:
                return ""
            else:
                return f"{c: .{precision}f} {name}"

        equations = []
        for coef_row in self.optimizer.coef_:
            components = [term(c, i) for c, i in zip(coef_row, feat_names)]
            eq = " + ".join(filter(bool, components))
            if not eq:
                eq = f"{0: .{precision}f}"
            equations.append(eq)

        return equations

    def print(self, precision: int = 3, **kwargs) -> None:
        """Print the SINDy model equations.

        Parameters
        ----------
        lhs: list of strings, optional (default None)
            List of variables to print on the left-hand sides of the learned equations.
            By default :code:`self.input_features` are used.

        precision: int, optional (default 3)
            Precision to be used when printing out model coefficients.

        **kwargs: Additional keyword arguments passed to the builtin print function
        """
        eqns = self.equations(precision)
        for name, eqn in zip(self.feature_names, eqns, strict=True):
            lhs = f"({name})'"
            print(f"{lhs} = {eqn}", **kwargs)

    def get_feature_names(self) -> list[str]:
        """
        Get a list of names of features used by SINDy model.

        Returns
        -------
        feats: list
            A list of strings giving the names of the features in the feature
            library, :code:`self.feature_library`.
        """
        check_is_fitted(self)
        return self.feature_library.get_feature_names(input_features=self.feature_names)


class SINDy(_BaseSINDy):
    """
    Sparse Identification of Nonlinear Dynamical Systems (SINDy).
    Uses sparse regression to learn a dynamical systems model from measurement data.

    Parameters
    ----------
    optimizer
        Optimization method used to fit the SINDy model. This must be a class
        extending :class:`pysindy.optimizers.BaseOptimizer`.
        The default is :class:`pysindy.optimizers.STLSQ`.

    feature_library
        Feature library object used to specify candidate right-hand side features.
        This must be a class extending
        :class:`pysindy.feature_library.base.BaseFeatureLibrary`.
        The default option is :class:`pysindy.feature_library.PolynomialLibrary`.

    differentiation_method
        Method for differentiating the data. This must be a class extending
        :class:`pysindy.differentiation.base.BaseDifferentiation` class.
        The default option is centered difference.

    Attributes
    ----------
    model : ``sklearn.multioutput.MultiOutputRegressor``
        The fitted SINDy model.

    n_input_features_ : int
        The total number of input features.

    n_output_features_ : int
        The total number of output features. This number is a function of
        ``self.n_input_features`` and the feature library being used.

    n_control_features_ : int
        The total number of control input features.

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.integrate import solve_ivp
    >>> from pysindy import SINDy
    >>> lorenz = lambda z,t : [10*(z[1] - z[0]),
    >>>                        z[0]*(28 - z[2]) - z[1],
    >>>                        z[0]*z[1] - 8/3*z[2]]
    >>> t = np.arange(0,2,.002)
    >>> x = solve_ivp(lorenz, [-8,8,27], t)
    >>> model = SINDy()
    >>> model.fit(x, t=t[1]-t[0])
    >>> model.print()
    x0' = -10.000 1 + 10.000 x0
    x1' = 27.993 1 + -0.999 x0 + -1.000 1 x1
    x2' = -2.666 x1 + 1.000 1 x0
    >>> model.coefficients()
    array([[ 0.        ,  0.        ,  0.        ],
           [-9.99969193, 27.99344519,  0.        ],
           [ 9.99961547, -0.99905338,  0.        ],
           [ 0.        ,  0.        , -2.66645651],
           [ 0.        ,  0.        ,  0.        ],
           [ 0.        ,  0.        ,  0.99990257],
           [ 0.        , -0.99980268,  0.        ],
           [ 0.        ,  0.        ,  0.        ],
           [ 0.        ,  0.        ,  0.        ],
           [ 0.        ,  0.        ,  0.        ]])
    >>> model.score(x, t=t[1]-t[0])
    0.999999985520653

    >>> import numpy as np
    >>> from scipy.integrate import solve_ivp
    >>> from pysindy import SINDy
    >>> u = lambda t : np.sin(2 * t)
    >>> lorenz_c = lambda z,t : [
                10 * (z[1] - z[0]) + u(t) ** 2,
                z[0] * (28 - z[2]) - z[1],
                z[0] * z[1] - 8 / 3 * z[2],
        ]
    >>> t = np.arange(0,2,0.002)
    >>> x = solve_ivp(lorenz_c, [-8,8,27], t)
    >>> u_eval = u(t)
    >>> model = SINDy()
    >>> model.fit(x, u_eval, t=t[1]-t[0])
    >>> model.print()
    x0' = -10.000 x0 + 10.000 x1 + 1.001 u0^2
    x1' = 27.994 x0 + -0.999 x1 + -1.000 x0 x2
    x2' = -2.666 x2 + 1.000 x0 x1
    >>> model.coefficients()
    array([[ 0.        , -9.99969851,  9.99958359,  0.        ,  0.        ,
             0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ,  0.        ,  0.        ,  0.        ,  1.00120331],
           [ 0.        , 27.9935177 , -0.99906375,  0.        ,  0.        ,
             0.        ,  0.        , -0.99980455,  0.        ,  0.        ,
             0.        ,  0.        ,  0.        ,  0.        ,  0.        ],
           [ 0.        ,  0.        ,  0.        , -2.666437  ,  0.        ,
             0.        ,  0.99990137,  0.        ,  0.        ,  0.        ,
             0.        ,  0.        ,  0.        ,  0.        ,  0.        ]])
    >>> model.score(x, u_eval, t=t[1]-t[0])
    0.9999999855414495
    """

    def __init__(
        self,
        optimizer: Optional[BaseOptimizer] = None,
        feature_library: Optional[BaseFeatureLibrary] = None,
        differentiation_method: Optional[BaseDifferentiation] = None,
    ):
        if optimizer is None:
            optimizer = STLSQ()
        self.optimizer = optimizer
        if feature_library is None:
            feature_library = PolynomialLibrary()
        self.feature_library = feature_library
        if differentiation_method is None:
            differentiation_method = FiniteDifference(axis=-2)
        self.differentiation_method = differentiation_method

    def fit(
        self,
        x,
        t,
        x_dot=None,
        u=None,
        feature_names: Optional[list[str]] = None,
    ):
        """
        Fit a SINDy model.

        Parameters
        ----------
        x: array-like or list of array-like, shape (n_samples, n_input_features)
            Training data. If training data contains multiple trajectories,
            x should be a list containing data for each trajectory. Individual
            trajectories may contain different numbers of samples.

        t: float, numpy array of shape (n_samples,), or list of numpy arrays
            If t is a float, it specifies the timestep between each sample.
            If array-like, it specifies the time at which each sample was
            collected.
            In this case the values in t must be strictly increasing.
            In the case of multi-trajectory training data, t may also be a list
            of arrays containing the collection times for each individual
            trajectory.

        x_dot: array-like or list of array-like, shape (n_samples, n_input_features), \
                optional (default None)
            Optional pre-computed derivatives of the training data. If not
            provided, the time derivatives of the training data will be
            computed using the specified differentiation method. If x_dot is
            provided, it must match the shape of the training data and these
            values will be used as the time derivatives.

        u: array-like or list of array-like, shape (n_samples, n_control_features), \
                optional (default None)
            Control variables/inputs. Include this variable to use sparse
            identification for nonlinear dynamical systems for control (SINDYc).
            If training data contains multiple trajectories (i.e. if x is a list of
            array-like), then u should be a list containing control variable data
            for each trajectory. Individual trajectories may contain different
            numbers of samples.

        feature_names : list of string, length n_input_features, optional
            Names for the input features (e.g. :code:`['x', 'y', 'z']`).
            If None, will use :code:`['x0', 'x1', ...]`.

        Returns
        -------
        self: a fitted :class:`SINDy` instance
        """

        if not _check_multiple_trajectories(x, x_dot, u):
            x, t, x_dot, u = _adapt_to_multiple_trajectories(x, t, x_dot, u)
        x, x_dot, u = _comprehend_and_validate_inputs(
            x, t, x_dot, u, self.feature_library
        )

        if x_dot is None:
            x, x_dot = self._process_trajectories(x, t, x_dot)

        if u is None:
            self.n_control_features_ = 0
        else:
            u = validate_control_variables(x, u)
            self.n_control_features_ = u[0].n_coord

            x = [np.concatenate((xi, ui), axis=xi.ax_coord) for xi, ui in zip(x, u)]

        self.feature_names = feature_names

        x_dot = concat_sample_axis(x_dot)
        x = self.feature_library.fit_transform(x)
        x = SampleConcatter().fit_transform(x)
        self.optimizer.fit(x, x_dot)
        self._fit_shape()

        return self

    def print(self, lhs=None, precision=3, **kwargs):
        """Print the SINDy model equations.

        Parameters
        ----------
        lhs: list of strings, optional (default None)
            List of variables to print on the left-hand sides of the learned equations.
            By default :code:`self.input_features` are used.

        precision: int, optional (default 3)
            Precision to be used when printing out model coefficients.

        **kwargs: Additional keyword arguments passed to the builtin print function
        """
        eqns = self.equations(precision)
        if sindy_pi_flag and isinstance(self.optimizer, SINDyPI):
            feature_names = self.get_feature_names()
        else:
            feature_names = self.feature_names
        for i, eqn in enumerate(eqns):
            if lhs is None:
                if not sindy_pi_flag or not isinstance(self.optimizer, SINDyPI):
                    names = f"({feature_names[i]})'"
                else:
                    names = f"({feature_names[i]})"
            else:
                names = f"{lhs[i]}"
            print(f"{names} = {eqn}", **kwargs)

    def score(self, x, t, x_dot=None, u=None, metric=r2_score, **metric_kws):
        """
        Returns a score for the time derivative prediction produced by the model.

        Parameters
        ----------
        x: array-like or list of array-like, shape (n_samples, n_input_features)
            Samples from which to make predictions.

        t: float, numpy array of shape (n_samples,), or list of numpy arrays
            Time step between samples or array of collection times. Optional,
            used to compute the time derivatives of the samples if x_dot is not
            provided.

        x_dot: array-like or list of array-like, shape (n_samples, n_input_features), \
                optional (default None)
            Optional pre-computed derivatives of the samples. If provided,
            these values will be used to compute the score. If not provided,
            the time derivatives of the training data will be computed using
            the specified differentiation method.

        u: array-like or list of array-like, shape(n_samples, n_control_features), \
                optional (default None)
            Control variables. If ``multiple_trajectories==True`` then u
            must be a list of control variable data from each trajectory.
            If the model was fit with control variables then u is not optional.

        metric: callable, optional
            Metric function with which to score the prediction. Default is the
            R^2 coefficient of determination.
            See `Scikit-learn \
            <https://scikit-learn.org/stable/modules/model_evaluation.html>`_
            for more options.

        metric_kws: dict, optional
            Optional keyword arguments to pass to the metric function.

        Returns
        -------
        score: float
            Metric function value for the model prediction of x_dot.
        """

        if not _check_multiple_trajectories(x, x_dot, u):
            x, t, x_dot, u = _adapt_to_multiple_trajectories(x, t, x_dot, u)
        x, x_dot, u = _comprehend_and_validate_inputs(
            x, t, x_dot, u, self.feature_library
        )

        x_dot_predict = self.predict(x, u)

        if x_dot is None:
            x, x_dot = self._process_trajectories(x, t, x_dot)

        x_dot = concat_sample_axis(x_dot)
        x_dot_predict = concat_sample_axis(x_dot_predict)

        x_dot, x_dot_predict = drop_nan_samples(x_dot, x_dot_predict)
        return metric(x_dot, x_dot_predict, **metric_kws)

    def _process_trajectories(self, x, t, x_dot):
        """
        Calculate derivatives of input data, iterating through trajectories.

        Parameters
        ----------
        x: list of np.ndarray
            List of measurements, with each entry corresponding to a different
            trajectory.

        t: list of np.ndarray or int
            List of time points for different trajectories.  If a list of ints
            is passed, each entry is assumed to be the timestep for the
            corresponding trajectory in x.  If np.ndarray is passed, it is
            used for each trajectory.

        x_dot: list of np.ndarray
            List of derivative measurements, with each entry corresponding to a
            different trajectory. If None, the derivatives will be approximated
            from x.

        Returns
        -------
        x_out: np.ndarray or list
            Validated version of x.

        x_dot_out: np.ndarray or list
            Validated derivative measurements
        """
        x, x_dot = zip(
            *[
                self.feature_library.calc_trajectory(
                    self.differentiation_method, xi, ti
                )
                for xi, ti in _zip_like_sequence(x, t)
            ]
        )
        return x, x_dot

    def simulate(
        self,
        x0,
        t,
        u=None,
        integrator="solve_ivp",
        interpolator=None,
        integrator_kws={"method": "LSODA", "rtol": 1e-12, "atol": 1e-12},
        interpolator_kws={},
    ):
        """
        Simulate the SINDy model forward in time.

        Parameters
        ----------
        x0: numpy array, size [n_features]
            Initial condition from which to simulate.

        t: numpy array of size [n_samples]
            Array of time points at which to simulate.

        u: function from R^1 to R^{n_control_features} or list/array, optional \
            (default None)
            Control inputs.
            ``u`` can be a function that would take time as input and output
            the values of each of the n_control_features control features as
            a list or numpy array. Alternatively, ``u`` can also be an array
            of control inputs at each time step. In this case, the array is fit
            with the interpolator specified by ``interpolator``.

        integrator: string, optional (default ``solve_ivp``)
            Function to use to integrate the system.
            Default is ``scipy.integrate.solve_ivp``. The only options
            currently supported are solve_ivp and odeint.

        interpolator: callable, optional (default ``interp1d``)
            Function used to interpolate control inputs if ``u`` is an array.
            Default is ``scipy.interpolate.interp1d``.

        integrator_kws: dict, optional (default {})
            Optional keyword arguments to pass to the integrator

        interpolator_kws: dict, optional (default {})
            Optional keyword arguments to pass to the control input interpolator

        Returns
        -------
        x: numpy array, shape (n_samples, n_features)
            Simulation results
        """
        check_is_fitted(self)
        if u is None and self.n_control_features_ > 0:
            raise TypeError("Model was fit using control variables, so u is required")

        if np.isscalar(t):
            raise ValueError(
                "For continuous time model, t must be an array of time"
                " points at which to simulate"
            )

        if u is None or self.n_control_features_ == 0:
            if u is not None:
                warnings.warn(
                    "Control variables u were ignored because control "
                    "variables were not used when the model was fit"
                )

            def rhs(t, x):
                return self.predict(x[np.newaxis, :])[0]

        else:
            if not callable(u):
                if interpolator is None:
                    u_fun = interp1d(
                        t, u, axis=0, kind="cubic", fill_value="extrapolate"
                    )
                else:
                    u_fun = interpolator(t, u, **interpolator_kws)

                t = t[:-1]
                warnings.warn(
                    "Last time point dropped in simulation because "
                    "interpolation of control input was used. To avoid "
                    "this, pass in a callable for u."
                )
            else:
                u_fun = u

            if u_fun(t[0]).ndim == 1:

                def rhs(t, x):
                    return self.predict(x[np.newaxis, :], u_fun(t).reshape(1, -1))[0]

            else:

                def rhs(t, x):
                    return self.predict(x[np.newaxis, :], u_fun(t))[0]

        # Need to hard-code below, because odeint and solve_ivp
        # have different syntax and integration options.
        if integrator == "solve_ivp":
            return ((solve_ivp(rhs, (t[0], t[-1]), x0, t_eval=t, **integrator_kws)).y).T
        elif integrator == "odeint":
            if integrator_kws.get("method") == "LSODA":
                integrator_kws = {}
            return odeint(rhs, x0, t, tfirst=True, **integrator_kws)
        else:
            raise ValueError("Integrator not supported, exiting")

    @property
    def complexity(self):
        """
        Complexity of the model measured as the number of nonzero parameters.
        """
        return self.optimizer.complexity


class DiscreteSINDy(_BaseSINDy):
    """
    Sparse Identification of Nonlinear Dynamical Systems (SINDy) for discrete
    time systems.

    Parameters
    ----------
    optimizer : optimizer object, optional
        Optimization method used to fit the SINDy model. This must be a class
        extending :class:`pysindy.optimizers.BaseOptimizer`.
        The default is :class:`STLSQ`.

    feature_library : feature library object, optional
        Feature library object used to specify candidate right-hand side features.
        This must be a class extending
        :class:`pysindy.feature_library.base.BaseFeatureLibrary`.
        The default option is :class:`PolynomialLibrary`.

    Attributes
    ----------
    model : ``sklearn.multioutput.MultiOutputRegressor`` object
        The fitted SINDy model.

    n_input_features_ : int
        The total number of input features.

    n_output_features_ : int
        The total number of output features. This number is a function of
        ``self.n_input_features`` and the feature library being used.

    n_control_features_ : int
        The total number of control input features.

    Examples
    --------
    >>> import numpy as np
    >>> import pysindy as ps
    >>> import matplotlib.pyplot as plt
    >>> def f(x):
    >>>     return 3.6 * x * (1 - x)
    >>> n_steps = 20
    >>> eps = 0.001
    >>> x_map = np.zeros((n_steps))
    >>> x_map[0] = 0.5
    >>> for i in range(1, n_steps):
    >>>     x_map[i] = f(x_map[i - 1]) + eps * np.random.randn()
    >>> x_train_map = x_map[:16]
    >>> x_test_map = x_map[16:]
    >>> model = ps.DiscreteSINDy()
    >>> model.fit(x_train_map, t=1)
    >>> model.print()
    (x0)[k+1] = 0.006 1 + 3.581 x0[k] + -3.586 x0[k]^2
    >>> model.coefficients()
    >>> model.predict(x_test_map)
    AxesArray([[0.8268863 ],
            [0.51884209],
            [0.89919392],
            [0.33532148]])
    >>> model.score(x_test_map, t=1)
    0.9998547755296847
    >>> model.simulate(x0=0.5, t=20)
    array([[0.5       ],
       [0.90037072],
       [0.32376537],
       [0.78980964],
       [0.59788467],
       [0.86556721],
       [0.41950949],
       [0.877508  ],
       [0.38763939],
       [0.85561537],
       [0.44528986],
       [0.88988815],
       [0.35351698],
       [0.8241012 ],
       [0.52224203],
       [0.89849516],
       [0.32914646],
       [0.79648209],
       [0.58382694],
       [0.87479097]])

    .. plot::
        :include-source: True

        num = 1000
        N = 1000
        N_drop = 500
        r0 = 3.5
        rs = r0 + np.arange(num) / num * (4 - r0)
        xss = []
        for r in rs:
            xs = []
            x = 0.5
            for n in range(N + N_drop):
                if n >= N_drop:
                    xs = xs + [x]
                x = r * x * (1 - x)
            xss = xss + [xs]
        plt.figure(figsize=(4, 4), dpi=100)
        for ind in range(num):
            plt.plot(
            np.ones(N) * rs[ind], xss[ind], ",", alpha=0.1, c="black", rasterized=True
            )
        plt.xlabel("$r$")
        plt.ylabel("$x_n$")
        plt.show()

    >>> rs_train = [3.6, 3.7, 3.8, 3.9]
    >>> xs_train = [np.array(xss[np.where(np.array(rs) == r)[0][0]]) for r in rs_train]
    >>> feature_lib = ps.PolynomialLibrary(degree=3, include_bias=True)
    >>> parameter_lib = ps.PolynomialLibrary(degree=1, include_bias=True)
    >>> lib = ps.ParameterizedLibrary(
    >>>     feature_library=feature_lib,
    >>>     parameter_library=parameter_lib,
    >>>     num_features=1,
    >>>     num_parameters=1,
    >>> )
    >>> opt = ps.STLSQ(threshold=1e-1, normalize_columns=False)
    >>> model = ps.DiscreteSINDy(feature_library=lib, optimizer=opt)
    >>> model.fit(xs_train, u=rs_train, t=1, feature_names=["x", "r"])
    >>> model.print()
    (x)[k+1] = 1.000 r[k] x[k] + -1.000 r[k] x[k]^2
    """

    def __init__(
        self,
        optimizer: Optional[BaseOptimizer] = None,
        feature_library: Optional[BaseFeatureLibrary] = None,
    ):
        if optimizer is None:
            optimizer = STLSQ()
        self.optimizer = optimizer
        if feature_library is None:
            feature_library = PolynomialLibrary()
        self.feature_library = feature_library

    def fit(
        self,
        x,
        t,
        x_next=None,
        u=None,
        feature_names: Optional[list[str]] = None,
    ):
        """
        Fit a DiscreteSINDy model.

        Parameters
        ----------
        x: array-like or list of array-like, shape (n_samples, n_input_features)
            Training data of the current state of the system. If training data
            contains multiple trajectories, x should be a list containing data for
            each trajectory. Individual trajectories may contain different numbers
            of samples.

        t: float, numpy array of shape (n_samples,), or list of numpy arrays
            If t is a float, it specifies the timestep between each sample.
            If array-like, it specifies the time at which each sample was
            collected.
            In this case the values in t must be strictly increasing.
            In the case of multi-trajectory training data, t may also be a list
            of arrays containing the collection times for each individual
            trajectory.

        x_next: array-like or list of array-like, shape (n_samples, n_input_features), \
                optional (default None)
            Optional data of the system forwarded by one time step. If not provided, the
            next will be computed by taking the training data by one time step.
            If x_next is provided, it must match the shape of the training data and
            these values will be used as the next state.

        u: array-like or list of array-like, shape (n_samples, n_control_features), \
                optional (default None)
            Control variables/inputs. Include this variable to use sparse
            identification for nonlinear dynamical systems for control (SINDYc).
            If training data contains multiple trajectories (i.e. if x is a list of
            array-like), then u should be a list containing control variable data
            for each trajectory. Individual trajectories may contain different
            numbers of samples.

        feature_names : list of string, length n_input_features, optional
            Names for the input features (e.g. :code:`['x', 'y', 'z']`).
            If None, will use :code:`['x0', 'x1', ...]`.

        Returns
        -------
        self: a fitted :class:`DiscreteSINDy` instance
        """

        if not _check_multiple_trajectories(x, x_next, u):
            x, t, x_next, u = _adapt_to_multiple_trajectories(x, t, x_next, u)
        x, x_next, u = _comprehend_and_validate_inputs(
            x, t, x_next, u, self.feature_library
        )

        if x_next is None:
            x_next = [xi[1:] for xi in x]
            x = [xi[:-1] for xi in x]
            if u is not None:
                u = [ui[:-1] for ui in u]

        # Append control variables
        if u is None:
            self.n_control_features_ = 0
        else:
            u = validate_control_variables(x, u)
            self.n_control_features_ = u[0].n_coord

            x = [np.concatenate((xi, ui), axis=xi.ax_coord) for xi, ui in zip(x, u)]

        self.feature_names = feature_names

        x_next = concat_sample_axis(x_next)
        x = self.feature_library.fit_transform(x)
        x = SampleConcatter().fit_transform(x)
        self.optimizer.fit(x, x_next)
        self._fit_shape()

        return self

    def equations(self, precision: int = 3) -> list[str]:
        """
        Get the right hand sides of the DiscreteSINDy model equations.

        Parameters
        ----------
        precision: int, optional (default 3)
            Number of decimal points to include for each coefficient in the
            equation.

        Returns
        -------
        equations: list of strings
            List of strings representing the DiscreteSINDy model equations for each
            input feature.
        """
        check_is_fitted(self)
        sys_coord_names = [name + "[k]" for name in self.feature_names]
        feat_names = self.feature_library.get_feature_names(sys_coord_names)

        def term(c, name):
            rounded_coef = np.round(c, precision)
            if rounded_coef == 0:
                return ""
            else:
                return f"{c: .{precision}f} {name}"

        equations = []
        for coef_row in self.optimizer.coef_:
            components = [term(c, i) for c, i in zip(coef_row, feat_names)]
            eq = " + ".join(filter(bool, components))
            if not eq:
                eq = f"{0: .{precision}f}"
            equations.append(eq)

        return equations

    def print(self, precision=3, **kwargs):
        """Print the DiscreteSINDy model equations.

        Parameters
        ----------
        precision: int, optional (default 3)
            Precision to be used when printing out model coefficients.

        **kwargs: Additional keyword arguments passed to the builtin print function
        """
        eqns = self.equations(precision)
        feature_names = self.feature_names
        for i, eqn in enumerate(eqns):
            names = f"({feature_names[i]})[k+1]"
            print(f"{names} = {eqn}", **kwargs)

    def score(self, x, t, u=None, x_next=None, metric=r2_score, **metric_kws):
        """
        Returns a score for the next state prediction produced by the model.

        Parameters
        ----------
        x: array-like or list of array-like, shape (n_samples, n_input_features)
            Samples from which to make predictions.

        t: float, numpy array of shape (n_samples,), or list of numpy arrays
            Time step between samples or array of collection times. Optional,
            used to compute the time derivatives of the samples if x_dot is not
            provided.

        u: array-like or list of array-like, shape(n_samples, n_control_features), \
                optional (default None)
            Control variables. If ``multiple_trajectories==True`` then u
            must be a list of control variable data from each trajectory.
            If the model was fit with control variables then u is not optional.

        metric: callable, optional
            Metric function with which to score the prediction. Default is the
            R^2 coefficient of determination.
            See `Scikit-learn \
            <https://scikit-learn.org/stable/modules/model_evaluation.html>`_
            for more options.

        metric_kws: dict, optional
            Optional keyword arguments to pass to the metric function.

        Returns
        -------
        score: float
            Metric function value for the model prediction of x_next.
        """

        if not _check_multiple_trajectories(x, x_next, u):
            x, t, x_next, u = _adapt_to_multiple_trajectories(x, t, x_next, u)
        x, x_next, u = _comprehend_and_validate_inputs(
            x, t, x_next, u, self.feature_library
        )

        x_next_predict = self.predict(x, u)

        if x_next is None:
            x_next_predict = [xd[:-1] for xd in x_next_predict]
            x_next = [xi[1:] for xi in x]
            x = [xi[:-1] for xi in x]

        x_next = concat_sample_axis(x_next)
        x_next_predict = concat_sample_axis(x_next_predict)

        x_next, x_next_predict = drop_nan_samples(x_next, x_next_predict)
        return metric(x_next, x_next_predict, **metric_kws)

    def simulate(
        self,
        x0,
        t,
        u=None,
        stop_condition=None,
    ):
        """
        Simulate the DiscreteSINDy model forward in time.

        Parameters
        ----------
        x0: numpy array, size [n_features]
            Initial condition from which to simulate.

        t: int
            Number of time steps to predict.

        u: list/array, optional (default None)
            Control inputs.
            A list (with ``len(u) == t``) or array (with ``u.shape[0] == 1``)
            giving the control inputs at each step.

        stop_condition: function object, optional
            If model is in discrete time, optional function that gives a
            stopping condition for stepping the simulation forward.

        Returns
        -------
        x: numpy array, shape (n_samples, n_features)
            Simulation results
        """
        check_is_fitted(self)
        if u is None and self.n_control_features_ > 0:
            raise TypeError("Model was fit using control variables, so u is required")

        if not isinstance(t, int) or t <= 0:
            raise ValueError(" t must be an integer")

        if stop_condition is not None:

            def check_stop_condition(xi):
                return stop_condition(xi)

        else:

            def check_stop_condition(xi):
                pass

        x = np.zeros((t, self.n_features_in_ - self.n_control_features_))
        x[0] = x0

        if u is None or self.n_control_features_ == 0:
            if u is not None:
                warnings.warn(
                    "Control variables u were ignored because control "
                    "variables were not used when the model was fit"
                )
            for i in range(1, t):
                x[i] = self.predict(x[i - 1 : i])
                if check_stop_condition(x[i]):
                    return x[: i + 1]
        else:
            for i in range(1, t):
                x[i] = self.predict(x[i - 1 : i], u=u[i - 1, np.newaxis])
                if check_stop_condition(x[i]):
                    return x[: i + 1]
        return x


def _zip_like_sequence(x, t):
    """Create an iterable like zip(x, t), but works if t is scalar.

    If t is an array, it is repeated for each x
    """
    if isinstance(t, Sequence):
        return zip(x, t)
    else:
        return product(x, [t])


def _check_multiple_trajectories(x, x_dot, u) -> bool:
    """Determine if data contains multiple trajectories

    Args:
        x: Samples from which to make predictions.
        x_dot: Pre-computed derivatives of the samples.
        u: Control variables

    Returns:
        whether data has multiple trajectories

    Raises:
        TypeError if data contains a mix of single/multiple trajectories
        ValueError if either data different numbers of trajectories

    """
    SequenceOrNone = Union[Sequence, None]
    mixed_trajectories = (
        isinstance(x, Sequence)
        and (not isinstance(x_dot, SequenceOrNone) or not isinstance(u, SequenceOrNone))
        or isinstance(x_dot, Sequence)
        and not isinstance(x, Sequence)
        or isinstance(u, Sequence)
        and not isinstance(x, Sequence)
    )
    if mixed_trajectories:
        raise TypeError(
            "If x, x_dot, or u are a Sequence of trajectories, each must be a Sequence"
            " of trajectories or None."
        )
    if isinstance(x, Sequence):
        matching_lengths = (x_dot is None or len(x) == len(x_dot)) and (
            u is None or len(x) == len(u)
        )
        if not matching_lengths:
            raise ValueError("x, x_dot and/or u have mismatched number of trajectories")
        return True
    return False


def _adapt_to_multiple_trajectories(x, t, x_dot, u) -> tuple:
    """Adapt model data to that multiple_trajectories.

    Args:
        x: Samples from which to make predictions.
        t: Time step between samples or array of collection times.
        x_dot: Pre-computed derivatives of the samples.
        u: Control variables

    Returns:
        Tuple of updated x, t, x_dot, u
    """
    x = [x]
    # if t is not a dt
    if not isinstance(t, np.ScalarType):
        t = [t]
    if x_dot is not None:
        x_dot = [x_dot]
    if u is not None:
        u = [u]
    return x, t, x_dot, u


def _comprehend_and_validate_inputs(x, t, x_dot, u, feature_library):
    """Validate input types, reshape arrays, and label axes"""

    def comprehend_and_validate(arr, t):
        arr = AxesArray(arr, comprehend_axes(arr))
        arr = feature_library.correct_shape(arr)
        return validate_no_reshape(arr, t)

    x = [comprehend_and_validate(xi, ti) for xi, ti in _zip_like_sequence(x, t)]
    if x_dot is not None:
        x_dot = [
            comprehend_and_validate(xdoti, ti)
            for xdoti, ti in _zip_like_sequence(x_dot, t)
        ]
    if u is not None:
        reshape_control = False
        for i in range(len(x)):
            if len(x[i].shape) != len(np.array(u[i]).shape):
                reshape_control = True
        if reshape_control:
            try:
                shape = np.array(x[0].shape)
                shape[x[0].ax_coord] = -1
                u = [np.reshape(u[i], shape) for i in range(len(x))]
            except Exception:
                try:
                    if np.isscalar(u[0]):
                        shape[x[0].ax_coord] = 1
                    else:
                        shape[x[0].ax_coord] = len(u[0])
                    u = [np.broadcast_to(u[i], shape) for i in range(len(x))]
                except Exception:
                    raise (
                        ValueError(
                            "Could not reshape control input to match the input data."
                        )
                    )
        correct_shape = True
        for i in range(len(x)):
            for axis in range(x[i].ndim):
                if (
                    axis != x[i].ax_coord
                    and x[i].shape[axis] != np.array(u[i]).shape[axis]
                ):
                    correct_shape = False
        if not correct_shape:
            raise (
                ValueError("Could not reshape control input to match the input data.")
            )
        u = [comprehend_and_validate(ui, ti) for ui, ti in _zip_like_sequence(u, t)]
    return x, x_dot, u



import numpy as np

from .base import BaseDifferentiation


class SpectralDerivative(BaseDifferentiation):
    """Spectral derivatives.
    Assumes uniform grid, and utilizes FFT to approximate a derivative.
    Works well for derivatives in periodic dimensions.
    Equivalent to a maximal-order finite difference, but runs in O(NlogN).

    Parameters
    ----------

    d : int
        The order of derivative to take
    axis: int, optional (default 0)
        The axis to differentiate along

    Examples
    --------
    >>> import numpy as np
    >>> from pysindy.differentiation import SpectralDerivative
    >>> t = np.arange(0,1,0.1)
    >>> X = np.vstack((np.sin(t), np.cos(t))).T
    >>> sd = SpectralDerivative()
    >>> sd._differentiate(X, t)
    array([[ 6.28318531e+00,  2.69942771e-16],
       [ 5.08320369e+00, -3.69316366e+00],
       [ 1.94161104e+00, -5.97566433e+00],
       [-1.94161104e+00, -5.97566433e+00],
       [-5.08320369e+00, -3.69316366e+00],
       [-6.28318531e+00,  7.10542736e-16],
       [-5.08320369e+00,  3.69316366e+00],
       [-1.94161104e+00,  5.97566433e+00],
       [ 1.94161104e+00,  5.97566433e+00],
       [ 5.08320369e+00,  3.69316366e+00]])
    """

    def __init__(self, d=1, axis=0):
        self.d = d
        self.axis = axis

    def _differentiate(self, x, t):
        """
        Calculate a spectral derivative.
        """

        if not np.isscalar(t):
            t = t[1] - t[0]

        q = np.fft.fft(x, axis=self.axis)
        n = x.shape[self.axis]
        dims = np.ones(x.ndim, dtype=int)
        dims[self.axis] = n
        freqs = np.zeros(n, dtype=np.complex128)
        positives = np.arange(n // 2 + 1)
        negatives = np.setdiff1d(np.arange(n), positives)
        freqs[: n // 2 + 1] = positives * 2 * np.pi / (n * t)
        freqs[n // 2 + 1 :] = (negatives - n) * 2 * np.pi / (n * t)

        self.smoothed_x_ = x
        if x.dtype is complex:
            return np.fft.ifft(
                np.reshape(1j * freqs, dims) ** self.d * q, axis=self.axis
            )
        else:
            return np.fft.ifft(
                np.reshape(1j * freqs, dims) ** self.d * q, axis=self.axis
            ).real.astype(x.dtype)



import numpy as np
import pytest
from numpy.testing import assert_array_equal

from pysindy.utils import AxesArray
from pysindy.utils import get_prox
from pysindy.utils import get_regularization
from pysindy.utils import reorder_constraints
from pysindy.utils import validate_control_variables


def test_reorder_constraints_1D():
    n_feats = 3
    n_tgts = 2
    target_order = np.array(
        [f"t{i}f{j}" for i in range(n_tgts) for j in range(n_feats)]
    )
    feature_order = np.array(
        [f"t{i}f{j}" for j in range(n_feats) for i in range(n_tgts)]
    )

    result = reorder_constraints(target_order, n_feats, output_order="feature")
    np.testing.assert_array_equal(result.flatten(), feature_order)

    result = reorder_constraints(feature_order, n_feats, output_order="target")
    np.testing.assert_array_equal(result.flatten(), target_order)


def test_reorder_constraints_2D():
    n_feats = 3
    n_tgts = 2
    n_const = 2
    target_order = np.array(
        [
            [f"c{k}t{i}f{j}" for i in range(n_tgts) for j in range(n_feats)]
            for k in range(n_const)
        ]
    )
    feature_order = np.array(
        [
            [f"c{k}t{i}f{j}" for j in range(n_feats) for i in range(n_tgts)]
            for k in range(n_const)
        ]
    )

    result = reorder_constraints(target_order, n_feats, output_order="feature")
    np.testing.assert_array_equal(result, feature_order)

    result = reorder_constraints(feature_order, n_feats, output_order="target")
    np.testing.assert_array_equal(result, target_order)


def test_validate_controls():
    with pytest.raises(ValueError):
        validate_control_variables(1, [])
    with pytest.raises(ValueError):
        validate_control_variables([], 1)
    with pytest.raises(ValueError):
        validate_control_variables([], [1])
    arr = AxesArray(np.ones(4).reshape((2, 2)), axes={"ax_time": 0, "ax_coord": 1})
    with pytest.raises(ValueError):
        validate_control_variables([arr], [arr[:1]])


@pytest.mark.parametrize(
    ["regularization", "lam", "expected"],
    [
        ("l0", 2, 4),
        ("l1", 2, 14),
        ("l2", 2, 58),
        ("weighted_l0", np.array([[3, 2]]).T, 5),
        ("weighted_l1", np.array([[3, 2]]).T, 16),
        ("weighted_l2", np.array([[3, 2]]).T, 62),
    ],
)
def test_get_regularization(regularization, lam, expected):
    data = np.array([[-2, 5]]).T

    reg = get_regularization(regularization)
    result = reg(data, lam)
    assert result == expected


@pytest.mark.parametrize("regularization", ["l0", "l1", "l2"])
@pytest.mark.parametrize(
    "lam",
    [
        np.array([[1, 2]]),
        np.array([[1]]),
    ],
)
def test_get_prox_and_regularization_bad_shape(regularization, lam):
    data = np.array([[-2, 5]]).T
    reg = get_regularization(regularization)
    with pytest.raises(ValueError):
        reg(data, lam)
    prox = get_prox(regularization)
    with pytest.raises(ValueError):
        prox(data, lam)


@pytest.mark.parametrize("regularization", ["l0", "l1", "l2"])
def test_get_weighted_prox_and_regularization_bad_shape(regularization):
    lam = np.array([[1, 2]])
    data = np.array([[-2, 5]]).T
    reg = get_regularization(regularization)
    with pytest.raises(ValueError):
        reg(data, lam)
    prox = get_prox(regularization)
    with pytest.raises(ValueError):
        prox(data, lam)


@pytest.mark.parametrize(
    ["regularization", "lam", "expected"],
    [
        ("l0", 1, np.array([[2]])),
        ("l1", 0.5, np.array([[1.5]])),
        ("l2", 0.5, np.array([[1]])),
        ("weighted_l0", np.array([[1]]), np.array([[2]])),
        ("weighted_l1", np.array([[0.5]]), np.array([[1.5]])),
        ("weighted_l2", np.array([[0.5]]), np.array([[1]])),
    ],
)
def test_get_prox(regularization, lam, expected):
    data = np.array([[2]])

    prox = get_prox(regularization)
    result = prox(data, lam)
    assert_array_equal(result, expected)



from importlib.metadata import PackageNotFoundError
from importlib.metadata import version

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    pass

from . import differentiation
from . import feature_library
from . import optimizers
from . import deeptime
from . import utils

from ._core import SINDy
from ._core import DiscreteSINDy
from ._core import AxesArray
from .differentiation import BaseDifferentiation
from .differentiation import FiniteDifference
from .differentiation import SpectralDerivative
from .differentiation import SINDyDerivative
from .differentiation import SmoothedFiniteDifference
from .feature_library import ConcatLibrary
from .feature_library import TensoredLibrary
from .feature_library import GeneralizedLibrary
from .feature_library import CustomLibrary
from .feature_library import FourierLibrary
from .feature_library import IdentityLibrary
from .feature_library import PolynomialLibrary
from .feature_library import PDELibrary
from .feature_library import WeakPDELibrary
from .feature_library import SINDyPILibrary
from .feature_library import ParameterizedLibrary
from .optimizers import BaseOptimizer
from .optimizers import FROLS
from .optimizers import WrappedOptimizer
from .optimizers import SR3
from .optimizers import SSR
from .optimizers import STLSQ
from .optimizers import EnsembleOptimizer

try:
    from .optimizers import ConstrainedSR3
except (ImportError, NameError):
    pass
try:
    from .optimizers import MIOSR
except (ImportError, NameError):
    pass
try:
    from .optimizers import SINDyPI
except (ImportError, NameError):
    pass
try:
    from .optimizers import TrappingSR3
except (ImportError, NameError):
    pass
try:
    from .optimizers import StableLinearSR3
except (ImportError, NameError):
    pass
try:
    from .optimizers import SBR
except (ImportError, NameError):
    pass


__all__ = [
    "SINDy",
    "DiscreteSINDy",
    "differentiation",
    "feature_library",
    "optimizers",
    "deeptime",
    "utils",
]



"""
Unit tests for SINDy class.

Note: all tests should be encapsulated in functions with
names starting with "test_"

To run all tests for this package, navigate to the top-level
directory and execute the following command:
pytest

To run tests for just one file, run
pytest file_to_test.py

"""
import numpy as np
import pytest
from sklearn.exceptions import ConvergenceWarning
from sklearn.exceptions import NotFittedError
from sklearn.linear_model import ElasticNet
from sklearn.linear_model import Lasso
from sklearn.model_selection import RandomizedSearchCV
from sklearn.model_selection import TimeSeriesSplit
from sklearn.utils.validation import check_is_fitted

from pysindy import _core
from pysindy import DiscreteSINDy
from pysindy import SINDy
from pysindy.differentiation import SINDyDerivative
from pysindy.differentiation import SmoothedFiniteDifference
from pysindy.feature_library import FourierLibrary
from pysindy.feature_library import PDELibrary
from pysindy.feature_library import PolynomialLibrary
from pysindy.feature_library import WeakPDELibrary
from pysindy.optimizers import ConstrainedSR3
from pysindy.optimizers import SR3
from pysindy.optimizers import STLSQ
from pysindy.optimizers import WrappedOptimizer


def test_get_feature_names_len(data_lorenz):
    x, t = data_lorenz
    model = SINDy()

    with pytest.raises(NotFittedError):
        model.get_feature_names()

    model.fit(x, t)

    # Assumes default library is polynomial features of degree 2
    assert len(model.get_feature_names()) == 10


def test_not_fitted(data_1d):
    x, t = data_1d
    model = SINDy()

    with pytest.raises(NotFittedError):
        model.predict(x)
    with pytest.raises(NotFittedError):
        model.get_feature_names()
    with pytest.raises(NotFittedError):
        model.coefficients()
    with pytest.raises(NotFittedError):
        model.equations()
    with pytest.raises(NotFittedError):
        model.simulate(x[0], t)


def test_improper_shape_input(data_1d):
    x, t = data_1d

    # Ensure model successfully handles different data shapes
    model = SINDy()
    model.fit(x.flatten(), t)
    check_is_fitted(model)

    model = SINDy()
    model.fit(x.flatten(), t, x_dot=x.flatten())
    check_is_fitted(model)

    model = SINDy()
    model.fit(x, t, x_dot=x.flatten())
    check_is_fitted(model)


@pytest.mark.parametrize(
    "data",
    [
        pytest.lazy_fixture("data_1d"),
        pytest.lazy_fixture("data_lorenz"),
        pytest.lazy_fixture("data_1d_bad_shape"),
    ],
)
def test_mixed_inputs(data):
    x, t = data

    # Scalar t
    model = SINDy()
    model.fit(x, t=2)
    check_is_fitted(model)

    model = SINDy()
    model.fit(x, t, x_dot=x)
    check_is_fitted(model)


@pytest.mark.parametrize(
    "data", [pytest.lazy_fixture("data_1d"), pytest.lazy_fixture("data_lorenz")]
)
def test_bad_t(data):
    x, t = data
    model = SINDy()

    # Wrong type
    with pytest.raises(ValueError):
        model.fit(x, t="1")

    # Invalid value of t
    with pytest.raises(ValueError):
        model.fit(x, t=-1)

    # t is a list
    with pytest.raises(ValueError):
        model.fit(x, list(t))

    # Wrong number of time points
    with pytest.raises(ValueError):
        model.fit(x, t[:-1])

    # Two points in t out of order
    t_new = np.copy(t)
    t_new[2], t_new[4] = t_new[4], t_new[2]
    with pytest.raises(ValueError):
        model.fit(x, t_new)
    t_new[2], t_new[4] = t_new[4], t_new[2]

    # Two matching times in t
    t_new[3] = t_new[5]
    with pytest.raises(ValueError):
        model.fit(x, t_new)


@pytest.mark.parametrize(
    "data", [pytest.lazy_fixture("data_1d"), pytest.lazy_fixture("data_lorenz")]
)
@pytest.mark.parametrize(
    "optimizer",
    [
        STLSQ(),
        SR3(),
        ConstrainedSR3(),
        WrappedOptimizer(Lasso(fit_intercept=False)),
        WrappedOptimizer(ElasticNet(fit_intercept=False)),
    ],
)
def test_predict(data, optimizer):
    x, t = data
    model = SINDy(optimizer=optimizer)
    model.fit(x, t)
    x_dot = model.predict(x)

    assert x.shape == x_dot.shape


@pytest.mark.parametrize(
    "data",
    [
        pytest.lazy_fixture("data_1d"),
        pytest.lazy_fixture("data_lorenz"),
        pytest.lazy_fixture("data_1d_bad_shape"),
    ],
)
def test_simulate(data):
    x, t = data
    model = SINDy(feature_library=PolynomialLibrary(degree=1))
    model.fit(x, t)
    x1 = model.simulate(np.ravel(x[0]), t, integrator_kws={"rtol": 0.1})
    assert len(x1) == len(t)
    x1 = model.simulate(
        np.ravel(x[0]), t, integrator="odeint", integrator_kws={"rtol": 0.1}
    )
    assert len(x1) == len(t)
    with pytest.raises(ValueError):
        x1 = model.simulate(np.ravel(x[0]), t, integrator="None")


@pytest.mark.parametrize(
    "library",
    [
        PolynomialLibrary(degree=3),
        FourierLibrary(n_frequencies=3),
        pytest.lazy_fixture("custom_library"),
        pytest.lazy_fixture("sindypi_library"),
        PolynomialLibrary() + FourierLibrary(),
    ],
)
def test_libraries(data_lorenz, library):
    x, t = data_lorenz
    model = SINDy(feature_library=library)
    model.fit(x, t)

    s = model.score(x, t)
    assert s <= 1


def test_integration_smoothed_finite_difference(data_lorenz):
    x, t = data_lorenz
    model = SINDy(differentiation_method=SmoothedFiniteDifference())

    model.fit(x, t=t)

    check_is_fitted(model)


@pytest.mark.parametrize(
    "derivative_kws",
    [
        dict(kind="finite_difference", k=1),
        dict(kind="spectral"),
        dict(kind="spline", s=1e-2),
        dict(kind="trend_filtered", order=0, alpha=1e-2),
        dict(kind="savitzky_golay", order=3, left=1, right=1),
    ],
)
def test_integration_derivative_methods(data_lorenz, derivative_kws):
    x, t = data_lorenz
    model = SINDy(differentiation_method=SINDyDerivative(**derivative_kws))
    model.fit(x, t=t)

    check_is_fitted(model)


@pytest.mark.parametrize(
    "data",
    [
        pytest.lazy_fixture("data_1d"),
        pytest.lazy_fixture("data_lorenz"),
        pytest.lazy_fixture("data_1d_bad_shape"),
    ],
)
def test_score(data):
    x, t = data
    model = SINDy()
    model.fit(x, t)

    assert model.score(x, t) <= 1

    assert model.score(x, t, x_dot=x) <= 1


def test_score_pde(data_1d_random_pde):
    t, x, u, u_dot = data_1d_random_pde
    pde_lib = PDELibrary(
        function_library=PolynomialLibrary(degree=2, include_bias=False),
        derivative_order=4,
        spatial_grid=x,
        include_bias=True,
    )
    model = SINDy(feature_library=pde_lib).fit(
        u,
        t,
    )

    assert model.score(u, t) <= 1

    assert model.score(u, t, x_dot=u) <= 1

    X, T = np.meshgrid(x, t)
    XT = np.array([X, T]).T
    weak_lib = WeakPDELibrary(
        function_library=PolynomialLibrary(degree=2, include_bias=False),
        derivative_order=4,
        spatiotemporal_grid=XT,
        include_bias=True,
    )
    model = SINDy(feature_library=weak_lib).fit(
        u,
        t=t,
    )

    assert model.score(u, t) <= 1


def test_fit_multiple_trajectories(data_multiple_trajectories):
    x, t = data_multiple_trajectories
    model = SINDy()

    model.fit(x, t=t)
    check_is_fitted(model)
    assert model.score(x, t=t) > 0.8

    model = SINDy()
    model.fit(x, t=t, x_dot=x)
    check_is_fitted(model)

    # Test validate_input
    t[0] = None
    with pytest.raises(ValueError):
        model.fit(x, t=t)


def test_predict_multiple_trajectories(data_multiple_trajectories):
    x, t = data_multiple_trajectories
    model = SINDy()
    model.fit(x, t=t)

    p = model.predict(x)
    assert len(p) == len(x)


def test_score_multiple_trajectories(data_multiple_trajectories):
    x, t = data_multiple_trajectories
    model = SINDy()
    model.fit(x, t=t)

    s = model.score(x, t=t)
    assert s <= 1

    s = model.score(x, t=t, x_dot=x)
    assert s <= 1


def test_fit_discrete_time(data_discrete_time):
    x = data_discrete_time

    model = DiscreteSINDy()
    model.fit(x, t=1)
    check_is_fitted(model)

    model = DiscreteSINDy()
    model.fit(x[:-1], x_next=x[1:], t=1)
    check_is_fitted(model)


def test_simulate_discrete_time(data_discrete_time):
    x = data_discrete_time
    model = DiscreteSINDy()
    model.fit(x, t=1)
    n_steps = x.shape[0]
    x1 = model.simulate(x[0], n_steps)

    assert len(x1) == n_steps

    def stop_func(xi):
        # check if we are at the 2nd to last element
        return np.isclose(xi[0], 0.874363)

    x2 = model.simulate(x[0], n_steps, stop_condition=stop_func)
    assert len(x2) == n_steps - 2


def test_predict_discrete_time(data_discrete_time):
    x = data_discrete_time
    model = DiscreteSINDy()
    model.fit(x, t=1)
    assert len(model.predict(x)) == len(x)


def test_score_discrete_time(data_discrete_time):
    x = data_discrete_time
    model = DiscreteSINDy()
    model.fit(x, t=1)
    assert model.score(x, t=1) > 0.75
    assert model.score(x, x_next=x, t=1) < 1


def test_bad_multiple_trajectories(data_multiple_trajectories):
    x, t = data_multiple_trajectories
    with pytest.raises(TypeError):
        _core._check_multiple_trajectories(x, x_dot=x[0], u=None)
    with pytest.raises(ValueError):
        _core._check_multiple_trajectories(x, x_dot=x[:-1], u=None)


def test_fit_discrete_time_multiple_trajectories(
    data_discrete_time_multiple_trajectories,
):
    x = data_discrete_time_multiple_trajectories
    model = DiscreteSINDy()
    model.fit(x, t=1)
    check_is_fitted(model)

    model = DiscreteSINDy()
    model.fit(x, x_next=x, t=1)
    check_is_fitted(model)


def test_predict_discrete_time_multiple_trajectories(
    data_discrete_time_multiple_trajectories,
):
    x = data_discrete_time_multiple_trajectories
    model = DiscreteSINDy()
    model.fit(x, t=1)

    y = model.predict(x)
    assert len(y) == len(x)


def test_score_discrete_time_multiple_trajectories(
    data_discrete_time_multiple_trajectories,
):
    x = data_discrete_time_multiple_trajectories
    model = DiscreteSINDy()
    model.fit(x, t=1)

    s = model.score(x, t=1)
    assert s > 0.75

    # x is not its own derivative, so we expect bad performance here
    s = model.score(x, x_next=x, t=1)
    assert s < 1


@pytest.mark.parametrize(
    "data",
    [
        pytest.lazy_fixture("data_1d"),
        pytest.lazy_fixture("data_lorenz"),
        pytest.lazy_fixture("data_1d_bad_shape"),
    ],
)
def test_equations(data, capsys):
    x, t = data
    model = SINDy()
    model.fit(x, t)

    out, _ = capsys.readouterr()
    assert len(out) == 0

    model.print(precision=2)

    out, _ = capsys.readouterr()

    assert len(out) > 0
    assert "(x0)' = " in out


def test_print_discrete_time(data_discrete_time, capsys):
    x = data_discrete_time
    model = DiscreteSINDy()
    model.fit(x, t=1)
    model.print()

    out, _ = capsys.readouterr()

    assert len(out) > 0
    assert "(x0)[k+1] = " in out


def test_coefficients_equals_complexity(data_lorenz):
    x, t = data_lorenz
    model = SINDy()
    model.fit(x, t)
    c = model.coefficients()
    assert model.complexity == np.count_nonzero(c)
    assert model.complexity < 30


def test_simulate_errors(data_lorenz):
    x, t = data_lorenz
    model = SINDy()
    model.fit(x, t)

    with pytest.raises(ValueError):
        model.simulate(x[0], t=1)

    model = DiscreteSINDy()
    with pytest.raises(ValueError):
        model.simulate(x[0], t=[1, 2])

    model = DiscreteSINDy()
    with pytest.raises(ValueError):
        model.simulate(x[0], t=-1)

    model = DiscreteSINDy()
    with pytest.raises(ValueError):
        model.simulate(x[0], t=0.5)


@pytest.mark.parametrize(
    "params, warning",
    [({"threshold": 100}, UserWarning), ({"max_iter": 1}, ConvergenceWarning)],
)
def test_fit_warn(data_lorenz, params, warning):
    x, t = data_lorenz
    model = SINDy(optimizer=STLSQ(**params))

    with pytest.warns(warning):
        model.fit(x, t=t)


def test_cross_validation(data_lorenz):
    x, t = data_lorenz

    model = SINDy(differentiation_method=SINDyDerivative(kind="spline", s=1e-2))

    param_grid = {
        "optimizer__threshold": [0.01, 0.1],
        "differentiation_method__kwargs": [
            {"kind": "spline", "s": 1e-2},
            {"kind": "finite_difference", "k": 1},
        ],
        "feature_library__degree": [1, 2],
    }

    search = RandomizedSearchCV(
        model, param_grid, cv=TimeSeriesSplit(n_splits=3), n_iter=5
    )
    search.fit(x, t=t)
    check_is_fitted(search)


def test_linear_constraints(data_lorenz):
    x, t = data_lorenz

    library = PolynomialLibrary().fit(x)

    constraint_rhs = np.ones(2)
    constraint_lhs = np.zeros((2, x.shape[1] * library.n_output_features_))

    target_1, target_2 = 1, 3
    constraint_lhs[0, 3] = target_1
    constraint_lhs[1, library.n_output_features_] = target_2

    optimizer = ConstrainedSR3(
        constraint_lhs=constraint_lhs, constraint_rhs=constraint_rhs
    )
    model = SINDy(feature_library=library, optimizer=optimizer).fit(x, t)

    coeffs = model.coefficients()

    np.testing.assert_allclose(
        np.array([coeffs[0, 3], coeffs[1, 0]]), np.array([1 / target_1, 1 / target_2])
    )


def test_data_shapes():
    model = SINDy()
    n = 10
    x = np.ones(n)
    t = 1
    model.fit(x, t)
    x = np.ones((n, 2))
    model.fit(x, t)
    x = np.ones((n, n, 2))
    model.fit(x, t)
    x = np.ones((n, n, n, 2))
    model.fit(x, t)
    x = np.ones((n, n, n, n, 2))
    model.fit(x, t)


def test_diffusion_pde(diffuse_multiple_trajectories):
    t, x, u = diffuse_multiple_trajectories

    pde_lib = PDELibrary(
        function_library=PolynomialLibrary(degree=1, include_bias=False),
        derivative_order=2,
        spatial_grid=x,
    )

    X, T = np.meshgrid(x, t, indexing="ij")
    XT = np.transpose([X, T], [1, 2, 0])

    weak_lib = WeakPDELibrary(
        function_library=PolynomialLibrary(degree=1, include_bias=False),
        derivative_order=2,
        spatiotemporal_grid=XT,
        K=100,
    )

    optimizer = STLSQ(threshold=0.2, alpha=1e-5, normalize_columns=False)
    model = SINDy(feature_library=pde_lib, optimizer=optimizer)
    model.fit(u, t=t, feature_names=["u"])
    assert abs(model.coefficients()[0, -1] - 1) < 1e-1
    assert np.all(model.coefficients()[0, :-1] == 0)

    model = SINDy(feature_library=weak_lib, optimizer=optimizer)
    model.fit(u, t=t, feature_names=["u"])
    assert abs(model.coefficients()[0, -1] - 1) < 1e-1
    assert np.all(model.coefficients()[0, :-1] == 0)



from itertools import chain
from math import comb
from typing import Iterator
from typing import Tuple

import numpy as np
from numpy.typing import NDArray
from scipy import sparse
from sklearn.preprocessing import PolynomialFeatures
from sklearn.utils.validation import check_is_fitted

from ..utils import AxesArray
from ..utils import comprehend_axes
from ..utils import wrap_axes
from ..utils._axis_conventions import AX_COORD
from .base import BaseFeatureLibrary
from .base import x_sequence_or_item


class PolynomialLibrary(BaseFeatureLibrary, PolynomialFeatures):
    """Generate polynomial and interaction features.

    This is the same as :code:`sklearn.preprocessing.PolynomialFeatures`,
    but also adds the option to omit interaction features from the library.

    Parameters
    ----------
    degree : integer, optional (default 2)
        The degree of the polynomial features.

    include_interaction : boolean, optional (default True)
        Determines whether interaction features are produced.
        If false, features are all of the form ``x[i] ** k``.

    interaction_only : boolean, optional (default False)
        If true, only interaction features are produced: features that are
        products of at most ``degree`` *distinct* input features (so not
        ``x[1] ** 2``, ``x[0] * x[2] ** 3``, etc.).

    include_bias : boolean, optional (default True)
        If True (default), then include a bias column, the feature in which
        all polynomial powers are zero (i.e. a column of ones - acts as an
        intercept term in a linear model).

    order : str in {'C', 'F'}, optional (default 'C')
        Order of output array in the dense case. 'F' order is faster to
        compute, but may slow down subsequent estimators.

    Attributes
    ----------
    powers_ : array, shape (n_output_features, n_input_features)
        powers_[i, j] is the exponent of the jth input in the ith output.

    n_features_in_ : int
        The total number of input features.

    n_output_features_ : int
        The total number of output features. This number is computed by
        iterating over all appropriately sized combinations of input features.
    """

    def __init__(
        self,
        degree=2,
        include_interaction=True,
        interaction_only=False,
        include_bias=True,
        order="C",
    ):
        super().__init__(
            degree=degree,
            interaction_only=interaction_only,
            include_bias=include_bias,
            order=order,
        )
        self.include_interaction = include_interaction

    @staticmethod
    def _combinations(
        n_features: int,
        degree: int,
        include_interaction: bool,
        interaction_only: bool,
        include_bias: bool,
    ) -> Iterator[Tuple[int, ...]]:
        """
        Create selection tuples of input indexes for each polynomail term

        Selection tuple iterates the input indexes present in a single term.
        For example, (x+y+1)^2 would be in iterator of the tuples:
        (), (0,), (1,), (0, 0), (0, 1), (1, 1)
        1    x     y      x^2     x*y     y^2

        Order of terms is preserved regardless of include_interation.
        """
        if not include_interaction:
            return chain(
                [()] if include_bias else [],
                (
                    exponent * (feat_idx,)
                    for exponent in range(1, degree + 1)
                    for feat_idx in range(n_features)
                ),
            )
        return PolynomialFeatures._combinations(
            n_features=n_features,
            min_degree=int(not include_bias),
            max_degree=degree,
            interaction_only=interaction_only,
            include_bias=include_bias,
        )

    @property
    def powers_(self) -> NDArray[np.int_]:
        """
        The exponents of the polynomial as an array of shape
        (n_features_out, n_features_in), where each item is the exponent of the
        jth input variable in the ith polynomial term.
        """
        check_is_fitted(self)
        combinations = self._combinations(
            n_features=self.n_features_in_,
            degree=self.degree,
            include_interaction=self.include_interaction,
            interaction_only=self.interaction_only,
            include_bias=self.include_bias,
        )
        return np.vstack(
            [np.bincount(c, minlength=self.n_features_in_) for c in combinations]
        )

    def get_feature_names(self, input_features=None):
        """Return feature names for output features.

        Parameters
        ----------
        input_features : list of string, length n_features, optional
            String names for input features if available. By default,
            "x0", "x1", ... "xn_features" is used.

        Returns
        -------
        output_feature_names : list of string, length n_output_features
        """
        powers = self.powers_
        if input_features is None:
            input_features = ["x%d" % i for i in range(powers.shape[1])]
        feature_names = []
        for row in powers:
            inds = np.where(row)[0]
            if len(inds):
                name = " ".join(
                    (
                        "%s^%d" % (input_features[ind], exp)
                        if exp != 1
                        else input_features[ind]
                    )
                    for ind, exp in zip(inds, row[inds])
                )
            else:
                name = "1"
            feature_names.append(name)
        return feature_names

    @x_sequence_or_item
    def fit(self, x_full: list[AxesArray], y=None):
        """
        Compute number of output features.

        Parameters
        ----------
        x : array-like, shape (n_samples, n_features)
            The data.

        Returns
        -------
        self : instance
        """
        if self.degree < 0 or not isinstance(self.degree, int):
            raise ValueError("degree must be a nonnegative integer")
        if (not self.include_interaction) and self.interaction_only:
            raise ValueError(
                "Can't have include_interaction be False and interaction_only"
                " be True"
            )
        n_features = x_full[0].shape[AX_COORD]
        combinations = self._combinations(
            n_features,
            self.degree,
            self.include_interaction,
            self.interaction_only,
            self.include_bias,
        )
        self.n_features_in_ = n_features
        self.n_output_features_ = sum(1 for _ in combinations)
        return self

    @x_sequence_or_item
    def transform(self, x_full):
        """Transform data to polynomial features.

        Parameters
        ----------
        x_full : {array-like, sparse matrix} of shape (n_samples, n_features)
            The data to transform, row by row.

        Returns
        -------
        xp : np.ndarray or CSR/CSC sparse matrix,
                shape (n_samples, n_output_features)
            The matrix of features, where n_output_features is the number
            of polynomial features generated from the combination of inputs.
        """
        check_is_fitted(self)

        xp_full = []
        for x in x_full:
            if sparse.issparse(x) and x.format not in ["csr", "csc"]:
                # create new with correct sparse
                axes = comprehend_axes(x)
                x = x.asformat("csc")
                wrap_axes(axes, x)
            n_features = x.shape[AX_COORD]
            if n_features != self.n_features_in_:
                raise ValueError("x shape does not match training shape")

            combinations = self._combinations(
                n_features,
                self.degree,
                self.include_interaction,
                self.interaction_only,
                self.include_bias,
            )
            if sparse.isspmatrix(x):
                columns = []
                for combo in combinations:
                    if combo:
                        out_col = 1
                        for col_idx in combo:
                            out_col = x[..., col_idx].multiply(out_col)
                        columns.append(out_col)
                    else:
                        bias = sparse.csc_matrix(np.ones((x.shape[0], 1)))
                        columns.append(bias)
                xp = sparse.hstack(columns, dtype=x.dtype).tocsc()
            else:
                xp = AxesArray(
                    np.empty(
                        (*x.shape[:-1], self.n_output_features_),
                        dtype=x.dtype,
                        order=self.order,
                    ),
                    x.axes,
                )
                for i, combo in enumerate(combinations):
                    xp[..., i] = x[..., combo].prod(-1)
            xp_full = xp_full + [xp]
        return xp_full


def n_poly_features(
    n_in_feat: int,
    degree: int,
    include_bias: bool = False,
    include_interation: bool = True,
    interaction_only: bool = False,
) -> int:
    """Calculate number of polynomial features

    Args:
        n_in_feat: number of input features, e.g. 3 for x, y, z
        degree: polynomial degree, e.g. 2 for quadratic
        include_bias: whether to include a constant term
        include_interaction: whether to include terms mixing multiple inputs
        interaction_only: whether to omit terms of x_m * x_n^p for p > 1
    """
    if not include_interation and interaction_only:
        raise ValueError("Cannot set interaction only if include_interaction is False")
    n_feat = include_bias
    if not include_interation:
        return n_feat + n_in_feat * degree
    for deg in range(1, degree + 1):
        if interaction_only:
            n_feat += comb(n_in_feat, deg)
        else:
            n_feat += comb(n_in_feat + deg - 1, deg)
    return n_feat


def IdentityLibrary():
    """
    Generate an identity library which maps all input features to
    themselves. An alias for creating a degree-1 polynomial library
    with no constant term.
    """
    return PolynomialLibrary(degree=1, include_bias=False)



from itertools import combinations
from itertools import combinations_with_replacement as combinations_w_r

from numpy import empty
from numpy import ones
from numpy import shape
from sklearn.utils.validation import check_is_fitted

from ..utils import AxesArray
from ..utils import comprehend_axes
from .base import BaseFeatureLibrary
from .base import x_sequence_or_item


class CustomLibrary(BaseFeatureLibrary):
    """Generate a library with custom functions.

    Parameters
    ----------
    library_functions : list of mathematical functions
        Functions to include in the library. Default is to use same functions
        for all variables. Can also be used so that each variable has an
        associated library, in this case library_functions is shape
        (n_input_features, num_library_functions)

    function_names : list of functions, optional (default None)
        List of functions used to generate feature names for each library
        function. Each name function must take a string input (representing
        a variable name), and output a string depiction of the respective
        mathematical function applied to that variable. For example, if the
        first library function is sine, the name function might return
        :math:`\\sin(x)` given :math:`x` as input. The function_names list
        must be the same length as library_functions. If no list of function
        names is provided, defaults to using
        :math:`[ f_0(x),f_1(x), f_2(x), \\ldots ]`.

    interaction_only : boolean, optional (default True)
        Whether to omit self-interaction terms.
        If True, function evaulations of the form :math:`f(x,x)` and :math:`f(x,y,x)`
        will be omitted, but those of the form :math:`f(x,y)` and :math:`f(x,y,z)`
        will be included.
        If False, all combinations will be included.

    include_bias : boolean, optional (default False)
        If True (default), then include a bias column, the feature in which
        all polynomial powers are zero (i.e. a column of ones - acts as an
        intercept term in a linear model).
        This is hard to do with just lambda functions, because if the system
        is not 1D, lambdas will generate duplicates.

    Attributes
    ----------
    functions : list of functions
        Mathematical library functions to be applied to each input feature.

    function_names : list of functions
        Functions for generating string representations of each library
        function.

    n_features_in_ : int
        The total number of input features.

    n_output_features_ : int
        The total number of output features. The number of output features
        is the product of the number of library functions and the number of
        input features.

    Examples
    --------
    >>> import numpy as np
    >>> from pysindy.feature_library import CustomLibrary
    >>> x = np.array([[0.,-1],[1.,0.],[2.,-1.]])
    >>> functions = [lambda x : np.exp(x), lambda x,y : np.sin(x+y)]
    >>> lib = CustomLibrary(library_functions=functions).fit(x)
    >>> lib.transform(x)
    array([[ 1.        ,  0.36787944, -0.84147098],
           [ 2.71828183,  1.        ,  0.84147098],
           [ 7.3890561 ,  0.36787944,  0.84147098]])
    >>> lib.get_feature_names()
    ['f0(x0)', 'f0(x1)', 'f1(x0,x1)']
    """

    def __init__(
        self,
        library_functions,
        function_names=None,
        interaction_only=True,
        include_bias=False,
    ):
        self.functions = library_functions
        self.function_names = function_names
        if function_names and (
            shape(library_functions)[-1] != shape(function_names)[-1]
        ):
            raise ValueError(
                "library_functions and function_names must have the same"
                " number of elements"
            )
        self.include_bias = include_bias
        self.interaction_only = interaction_only

    @staticmethod
    def _combinations(n_features, n_args, interaction_only):
        """Get the combinations of features to be passed to a library function."""
        comb = combinations if interaction_only else combinations_w_r
        return comb(range(n_features), n_args)

    def get_feature_names(self, input_features=None):
        """Return feature names for output features.

        Parameters
        ----------
        input_features : list of string, length n_features, optional
            String names for input features if available. By default,
            "x0", "x1", ... "xn_features" is used.

        Returns
        -------
        output_feature_names : list of string, length n_output_features
        """
        check_is_fitted(self)

        n_input_features = self.n_features_in_
        if input_features is None:
            input_features = ["x%d" % i for i in range(n_input_features)]
        feature_names = []
        if self.include_bias:
            feature_names.append("1")
        for i, f in enumerate(self.functions):
            for c in self._combinations(
                n_input_features, f.__code__.co_argcount, self.interaction_only
            ):
                feature_names.append(
                    self.function_names[i](*[input_features[j] for j in c])
                )
        return feature_names

    @x_sequence_or_item
    def fit(self, x_full, y=None):
        """Compute number of output features.

        Parameters
        ----------
        x : array-like, shape (n_samples, n_features)
            Measurement data.

        Returns
        -------
        self : instance
        """
        n_features = x_full[0].shape[x_full[0].ax_coord]
        self.n_features_in_ = n_features
        n_output_features = 0
        for f in self.functions:
            n_args = f.__code__.co_argcount
            n_output_features += len(
                list(self._combinations(n_features, n_args, self.interaction_only))
            )
        if self.include_bias:
            n_output_features += 1
        self.n_output_features_ = n_output_features
        if self.function_names is None:
            self.function_names = list(
                map(
                    lambda i: (lambda *x: "f" + str(i) + "(" + ",".join(x) + ")"),
                    range(len(self.functions)),
                )
            )
        return self

    @x_sequence_or_item
    def transform(self, x_full):
        """Transform data to custom features

        Parameters
        ----------
        x : array-like, shape (n_samples, n_features)
            The data to transform, row by row.

        Returns
        -------
        xp : np.ndarray, shape (n_samples, n_output_features)
            The matrix of features, where n_output_features is the number of features
            generated from applying the custom functions to the inputs.
        """
        check_is_fitted(self)

        xp_full = []
        for x in x_full:
            n_features = x.shape[x.ax_coord]
            n_input_features = self.n_features_in_

            if n_features != n_input_features:
                raise ValueError("x shape does not match training shape")

            xp = empty((*x.shape[:-1], self.n_output_features_), dtype=x.dtype)
            library_idx = 0
            if self.include_bias:
                xp[..., library_idx] = ones(x.shape[:-1])
                library_idx += 1
            for f in self.functions:
                for c in self._combinations(
                    n_input_features, f.__code__.co_argcount, self.interaction_only
                ):
                    xp[..., library_idx] = f(*[x[..., j] for j in c])
                    library_idx += 1

            xp = AxesArray(xp, comprehend_axes(xp))
            xp_full.append(xp)
        return xp_full



import warnings

import cvxpy as cp
import numpy as np
from sklearn.exceptions import ConvergenceWarning

from .sr3 import SR3


class SINDyPI(SR3):
    """
    SINDy-PI optimizer

    Attempts to minimize the objective function

    .. math::

        0.5\\|X-Xw\\|^2_2 + \\lambda R(w)

    over w where :math:`R(v)` is a regularization function. See the following
    reference for more details:

        Kaheman, Kadierdan, J. Nathan Kutz, and Steven L. Brunton. SINDy-PI:
        a robust algorithm for parallel implicit sparse identification of
        nonlinear dynamics.
        Proceedings of the Royal Society A 476.2242 (2020): 20200279.

    Parameters
    ----------
    regularizer : string, optional (default 'l1')
        Regularization function to use. Currently implemented options
        are 'l1' (l1 norm), 'weighted_l1' (weighted l1 norm), l2, and
        'weighted_l2' (weighted l2 norm)

    model_subset : np.ndarray, shape(n_models), optional (default None)
        List of indices to compute models for. If list is not provided,
        the default is to compute SINDy-PI models for all possible
        candidate functions. This can take a long time for 4D systems
        or larger.

    verbose_cvxpy : bool, optional (default False)
        Boolean flag which is passed to CVXPY solve function to indicate if
        output should be verbose or not. Only relevant for optimizers that
        use the CVXPY package in some capabity.

    Attributes
    ----------
    coef_ : array, shape (n_features,) or (n_targets, n_features)
        Regularized weight vector(s). This is the v in the objective
        function.

    unbias: bool
        Required to be false, maintained for supertype compatibility
    """

    def __init__(
        self,
        reg_weight_lam=0.1,
        regularizer="l1",
        tol=1e-5,
        max_iter=10000,
        copy_X=True,
        model_subset=None,
        normalize_columns=False,
        verbose_cvxpy=False,
        unbias=False,
    ):
        super().__init__(
            reg_weight_lam=reg_weight_lam,
            regularizer=regularizer,
            tol=tol,
            max_iter=max_iter,
            copy_X=copy_X,
            normalize_columns=normalize_columns,
            unbias=unbias,
        )

        if regularizer.lower() not in ["l1", "l2", "weighted_l1", "weighted_l2"]:
            raise ValueError(
                "l0 and other nonconvex regularizers are not implemented "
                " in current version of SINDy-PI"
            )

        if self.unbias:
            raise ValueError("SINDyPI is incompatible with an unbiasing step")
        self.verbose_cvxpy = verbose_cvxpy
        if model_subset is not None:
            if not isinstance(model_subset, list):
                raise ValueError("Model subset must be in list format.")
            subset_integers = [
                model_ind for model_ind in model_subset if isinstance(model_ind, int)
            ]
            if subset_integers != model_subset:
                raise ValueError("Model subset list must consist of integers.")

        self.model_subset = model_subset

    def _update_parallel_coef_constraints(self, x):
        """
        Solves each of the model fits separately, which can in principle be
        parallelized. Unfortunately most parallel Python packages don't give
        huge speedups. Instead, we allow the user to only solve a subset of
        the models with the parameter model_subset.
        """
        n_features = x.shape[1]
        xi_final = np.zeros((n_features, n_features))

        # Todo: parallelize this for loop with Multiprocessing/joblib
        if self.model_subset is None:
            self.model_subset = range(n_features)
        elif np.max(np.abs(self.model_subset)) >= n_features:
            raise ValueError(
                "A value in model_subset is larger than the number "
                "of features in the candidate library"
            )
        for i in self.model_subset:
            print("Model ", i)
            xi = cp.Variable(n_features)
            # Note that norm choice below must be convex,
            # so regularizer must be L1 or L2
            regularizer = self.regularizer.lower()
            lam = self.reg_weight_lam
            if regularizer == "l1":
                cost = cp.sum_squares(x[:, i] - x @ xi) + cp.sum(
                    cp.multiply(lam, cp.abs(xi))
                )
            elif regularizer == "weighted_l1":
                cost = cp.sum_squares(x[:, i] - x @ xi) + cp.sum(
                    cp.multiply(lam[:, i], cp.abs(xi))
                )
            elif regularizer == "l2":
                cost = cp.sum_squares(x[:, i] - x @ xi) + cp.sum(
                    cp.multiply(lam, xi**2)
                )
            elif regularizer == "weighted_l2":
                cost = cp.sum_squares(x[:, i] - x @ xi) + cp.sum(
                    cp.multiply(lam[:, i], xi**2)
                )
            prob = cp.Problem(
                cp.Minimize(cost),
                [xi[i] == 0.0],
            )
            try:
                prob.solve(
                    max_iter=self.max_iter,
                    eps_abs=self.tol,
                    eps_rel=self.tol,
                    verbose=self.verbose_cvxpy,
                )
                if xi.value is None:
                    warnings.warn(
                        "Infeasible solve on iteration "
                        + str(i)
                        + ", try changing your library",
                        ConvergenceWarning,
                    )
                xi_final[:, i] = xi.value
            except cp.error.SolverError:
                print("Solver failed on model ", str(i), ", setting coefs to zeros")
                xi_final[:, i] = np.zeros(n_features)
        return xi_final

    def _reduce(self, x, y):
        """
        Perform at most ``self.max_iter`` iterations of the SINDy-PI
        optimization problem, using CVXPY.
        """
        coef = self._update_parallel_coef_constraints(x)
        self.coef_ = coef.T



import numpy as np

from .base import BaseFeatureLibrary
from .generalized_library import GeneralizedLibrary
from .polynomial_library import PolynomialLibrary


class ParameterizedLibrary(GeneralizedLibrary):
    """A tensor product of two libraries with different inputs.
    Typically, this is a feature library of the input data and a
    parameter library of input control, making the SINDyCP method.  If
    the input libraries are weak, the temporal derivatives are
    automatically rescaled by the appropriate domain volumes.

    Parameters
    ----------
    parameter_library : BaseFeatureLibrary, optional (default PolynomialLibrary).
    Specifies the library function to apply to the input control features.

    feature_library : BaseFeatureLibrary, optional (default PolynomialLibrary).
    Specifies the library function to apply to the input data features.

    num_parameters : int, optional (default 3)
    Specifies the number of features in the input control.

    num_features : int, optional (default 3)
    Specifies the number of features in the input data.

    Attributes
    ----------
    see GeneralizedLibrary

    Examples
    ----------
    >>> import numpy as np
    >>> from pysindy.feature_library import ParameterizedLibrary,PolynomialLibrary
    >>> from pysindy import AxesArray
    >>> xs=[np.random.random((5,3)) for n in range(3)]
    >>> us=[np.random.random((5,3)) for n in range(3)]
    >>> feature_lib=PolynomialLibrary(degree=3)
    >>> parameter_lib=PolynomialLibrary(degree=1)
    >>> lib=ParameterizedLibrary(feature_library=feature_lib,
    >>>     parameter_library=parameter_lib,num_features=3,num_parameters=3)
    >>> xus=[AxesArray(np.concatenate([xs[i],us[i]],axis=-1)) for i in range(3)]
    >>> lib.fit(xus)
    >>> lib.transform(xus)
    """

    def __init__(
        self,
        parameter_library: BaseFeatureLibrary = PolynomialLibrary(degree=1),
        feature_library: BaseFeatureLibrary = PolynomialLibrary(),
        num_parameters: int = 3,
        num_features: int = 3,
    ):
        if num_parameters <= 0 or num_features <= 0:
            raise ValueError("Both num_parameter and num_feature must be positive.")
        inputs_per_library = [
            range(num_features, num_parameters + num_features),
            range(num_features),
        ]

        super().__init__(
            libraries=[parameter_library, feature_library],
            tensor_array=[[1, 1]],
            exclude_libraries=[0, 1],
            inputs_per_library=inputs_per_library,
        )

    def calc_trajectory(self, diff_method, x, t):
        # if tensoring weak libraries, add the correction
        if hasattr(self.libraries[0], "K"):
            constants_final = np.ones(self.libraries[0].K)
            for k in range(self.libraries[0].K):
                constants_final[k] = np.sum(self.libraries[0].fullweights0[k])
            x, x_int = self.libraries[0].calc_trajectory(diff_method, x, t)
            return x, x_int * constants_final[:, np.newaxis]
        else:
            return self.libraries[0].calc_trajectory(diff_method, x, t)



"""
A module that defines one external class, AxesArray, to act like a numpy array
but keep track of axis definitions.  It aims to allow meaningful replacement
of magic numbers for axis conventions in code.  E.g::

   import numpy as np

   arr = AxesArray(np.ones((2,3,4)), {"ax_time": 0, "ax_spatial": [1, 2]})
   print(arr.axes)
   print(arr.ax_time)
   print(arr.n_time)
   print(arr.ax_spatial)
   print(arr.n_spatial)

Would show::

   {"ax_time": 0, "ax_spatial": [1, 2]}
   0
   2
   [1, 2]
   [3, 4]

It is up to the user to handle the ``list[int] | int`` return values, but this
module has several functions to deal with the axes dictionary, internally
referred to as type ``CompatDict[T]``:

Appending an item to a ``CompatDict[T]``
   :py:func:`compat_dict_append`

Generating a ``CompatDict[int]`` of axes from list of axes names:
   :py:func:`fwd_from_names`

Create new ``CompatDict[int]`` from this ``AxesArray`` with new axis/axes added:
   :py:meth:`AxesArray.insert_axis`

Create new ``CompatDict[int]`` from this ``AxesArray`` with axis/axes removed:
   :py:meth:`AxesArray.remove_axis`


.. todo::

   Add developer documentation here.

The recommended way to refactor existing code to use AxesArrays is to add them
at the lowest level possible.  Enter debug mode and see how long the expected
axes persist throughout array operations.  When AxesArray loses track of the
correct axes, re-assign them with an AxesArray constructor (which only uses a
view of the data).

Starting at the macro level runs the risk of triggering a great deal of errors
from unimplemented functions.
"""
from __future__ import annotations

import copy
import warnings
from enum import Enum
from typing import Collection
from typing import Dict
from typing import get_args
from typing import List
from typing import Literal
from typing import NewType
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import TypeVar
from typing import Union

import numpy as np
from numpy.typing import NDArray
from sklearn.base import TransformerMixin

HANDLED_FUNCTIONS = {}

AxesWarning = type("AxesWarning", (SyntaxWarning,), {})
BasicIndexer = Union[slice, int, type(Ellipsis), None, str]
Indexer = Union[BasicIndexer, NDArray, List]
StandardIndexer = Union[slice, int, None, NDArray[np.dtype(int)]]
OldIndex = NewType("OldIndex", int)  # Before moving advanced axes adajent
KeyIndex = NewType("KeyIndex", int)
NewIndex = NewType("NewIndex", int)
T = TypeVar("T", bound=int)  # TODO: Bind to a non-sequence after type-negation PEP
ItemOrList = Union[T, List[T]]
CompatDict = Dict[str, ItemOrList[T]]


class _Sentinels(Enum):
    ADV_NAME = object()
    ADV_REMOVE = object()


class _AxisMapping:
    """Convenience wrapper for a two-way map between axis names and indexes."""

    fwd_map: Dict[str, List[int]]
    reverse_map: Dict[int, str]

    def __init__(
        self,
        axes: dict[str, Union[int, Sequence[int]]],
        in_ndim: int,
    ):
        self.fwd_map = {}
        self.reverse_map = {}

        def coerce_sequence(obj):
            if isinstance(obj, Sequence):
                return sorted(obj)
            return [obj]

        for ax_name, ax_ids in axes.items():
            ax_ids = coerce_sequence(ax_ids)
            self.fwd_map[ax_name] = ax_ids
            for ax_id in ax_ids:
                old_name = self.reverse_map.get(ax_id)
                if old_name is not None:
                    raise ValueError(f"Assigned multiple definitions to axis {ax_id}")
                if ax_id >= in_ndim:
                    raise ValueError(
                        f"Assigned definition to axis {ax_id}, but array only has"
                        f" {in_ndim} axes"
                    )
                self.reverse_map[ax_id] = ax_name
        if len(self.reverse_map) != in_ndim:
            warnings.warn(
                f"{len(self.reverse_map)} axes labeled for array with {in_ndim} axes",
                AxesWarning,
            )

    @staticmethod
    def _compat_axes(in_dict: Dict[str, List[int]]) -> Dict[str, Union[list[int], int]]:
        """Like fwd_map, but unpack single-element axis lists"""
        axes = {}
        for k, v in in_dict.items():
            if len(v) == 1:
                axes[k] = v[0]
            else:
                axes[k] = v
        return axes

    @property
    def compat_axes(self):
        return self._compat_axes(self.fwd_map)

    def remove_axis(self, axis: Union[Collection[int], int, None] = None):
        """Create an axes dict from self with specified axis or axes
        removed and all greater axes decremented.  This can be passed to
        the constructor to create a new _AxisMapping

        Parameters:
            axis: the axis index or axes indexes to remove.  By numpy
            ufunc convention, axis=None (default) removes _all_ axes.
        """
        if axis is None:
            return {}
        new_axes = copy.deepcopy(self.fwd_map)
        in_ndim = len(self.reverse_map)
        if not isinstance(axis, Collection):
            axis = [axis]
        axis = [ax_id if ax_id >= 0 else (self.ndim + ax_id) for ax_id in axis]
        for cum_shift, orig_ax_remove in enumerate(sorted(axis)):
            remove_ax_name = self.reverse_map[orig_ax_remove]
            curr_ax_remove = orig_ax_remove - cum_shift
            if len(new_axes[remove_ax_name]) == 1:
                new_axes.pop(remove_ax_name)
            else:
                new_axes[remove_ax_name].remove(curr_ax_remove)
            for old_ax_dec in range(curr_ax_remove + 1, in_ndim - cum_shift):
                orig_ax_dec = old_ax_dec + cum_shift
                ax_dec_name = self.reverse_map[orig_ax_dec]
                new_axes[ax_dec_name].remove(old_ax_dec)
                new_axes[ax_dec_name].append(old_ax_dec - 1)
        return self._compat_axes(new_axes)

    def insert_axis(self, axis: Union[Collection[int], int], new_name: str):
        """Create an axes dict from self with specified axis or axes
        added and all greater axes incremented.

        Parameters:
            axis: the axis index or axes indexes to add.

        Todo:
            May be more efficient to determine final axis-to-axis
            mapping, then apply, rather than apply changes after each
            axis insert.
        """
        new_axes = copy.deepcopy(self.fwd_map)
        in_ndim = len(self.reverse_map)
        if not isinstance(axis, Collection):
            axis = [axis]
        for cum_shift, ax in enumerate(sorted(axis)):
            if new_name in new_axes.keys():
                new_axes[new_name].append(ax)
            else:
                new_axes[new_name] = [ax]
            for ax_id in range(ax, in_ndim + cum_shift):
                ax_name = self.reverse_map[ax_id - cum_shift]
                new_axes[ax_name].remove(ax_id)
                new_axes[ax_name].append(ax_id + 1)
        return self._compat_axes(new_axes)

    @property
    def ndim(self):
        return len(self.reverse_map)


class AxesArray(np.lib.mixins.NDArrayOperatorsMixin, np.ndarray):
    """A numpy-like array that keeps track of the meaning of its axes.

    Limitations:

    * Not all numpy functions, such as ``np.flatten()``, have an
      implementation for ``AxesArray``.  In such cases a regular numpy array
      is returned.
    * For functions that are implemented for ``AxesArray``, such as
      ``np.reshape()``, use the numpy function rather than the bound
      method (e.g. ``arr.reshape``)
    * Such functions may raise ``ValueError`` where numpy would not, when
      it is impossible to determine the output axis labels.

    Current array function implementations:

    * ``np.concatenate``
    * ``np.reshape``
    * ``np.transpose``
    * ``np.linalg.solve``
    * ``np.einsum``
    * ``np.tensordot``

    Indexing:
        AxesArray supports all of the basic and advanced indexing of numpy
        arrays, with the addition that new axes can be inserted with a string
        name for the axis.  E.g. ``arr = arr[..., "lineno"]`` will add a
        length-one axis at the end, along with the properties ``arr.ax_lineno``
        and ``arr.n_lineno``.  If ``None`` or ``np.newaxis`` are passed, the
        axis is named "unk".

    Parameters:
        input_array: the data to create the array.
        axes: A dictionary of axis labels to shape indices.  Axes labels must
            be of the format ``"ax_name"``.  indices can be either an int or a
            list of ints.

    Attributes:
        axes: dictionary of axis name to dimension index/indices
        ax_<ax_name>: lookup ax_name in axes
        n_<ax_name>: lookup shape of subarray defined by ax_name

    Raises:
        AxesWarning if axes does not match shape of input_array.
        ValueError if assigning the same axis index to multiple meanings or
            assigning an axis beyond ndim.

    """

    _ax_map: _AxisMapping

    def __new__(cls, input_array: NDArray, axes: CompatDict[int]):
        obj = np.asarray(input_array).view(cls)
        in_ndim = len(input_array.shape)
        obj._ax_map = _AxisMapping(axes, in_ndim)
        return obj

    @property
    def axes(self):
        return self._ax_map.compat_axes

    @property
    def _reverse_map(self):
        return self._ax_map.reverse_map

    @property
    def shape(self):
        """Shape of array.  Unlike numpy ndarray, this is not assignable."""
        return super().shape

    def insert_axis(
        self, axis: Union[Collection[int], int], new_name: str
    ) -> CompatDict[int]:
        """Create the constructor axes dict from this array, with new axis/axes"""
        return self._ax_map.insert_axis(axis, new_name)

    def remove_axis(self, axis: Union[Collection[int], int]) -> CompatDict[int]:
        """Create the constructor axes dict from this array, without axis/axes"""
        return self._ax_map.remove_axis(axis)

    def __getattr__(self, name):
        # TODO: replace with structural pattern matching on Oct 2025 (3.9 EOL)
        parts = name.split("_", 1)
        if parts[0] == "ax":
            try:
                return self.axes[name]
            except KeyError:
                raise AttributeError(f"AxesArray has no axis '{name}'")
        if parts[0] == "n":
            try:
                ax_ids = self._ax_map.fwd_map["ax_" + parts[1]]
            except KeyError:
                raise AttributeError(f"AxesArray has no axis '{name}'")
            shape = tuple(self.shape[ax_id] for ax_id in ax_ids)
            if len(shape) == 1:
                return shape[0]
            return shape
        raise AttributeError(f"'{type(self)}' object has no attribute '{name}'")

    def __getitem__(self, key: Union[Indexer, Sequence[Indexer]], /):
        if isinstance(key, tuple):
            base_indexer = tuple(None if isinstance(k, str) else k for k in key)
        else:
            base_indexer = key
        output = super().__getitem__(base_indexer)
        if not isinstance(output, AxesArray):
            return output  # return an element from the array
        in_dim = self.shape
        key, adv_inds = _standardize_indexer(self, key)
        bcast_nd, bcast_start_ax = _determine_adv_broadcasting(key, adv_inds)
        if adv_inds:
            key = _replace_adv_indexers(key, adv_inds, bcast_start_ax, bcast_nd)
        remove_axes, new_axes, adv_names = _apply_indexing(key, self._reverse_map)
        new_axes = _rename_broadcast_axes(new_axes, adv_names)
        new_map = _AxisMapping(
            self._ax_map.remove_axis(remove_axes), len(in_dim) - len(remove_axes)
        )
        for insert_counter, (new_ax_ind, new_ax_name) in enumerate(new_axes):
            new_map = _AxisMapping(
                new_map.insert_axis(new_ax_ind, new_ax_name),
                in_ndim=len(in_dim) - len(remove_axes) + (insert_counter + 1),
            )
        output._ax_map = new_map
        return output

    def __array_finalize__(self, obj) -> None:
        if obj is None:  # explicit construction via super().__new__()
            return
        # view from numpy array, called in constructor but also tests
        if all(
            (
                not isinstance(obj, AxesArray),
                self.shape == (),
                not hasattr(self, "_ax_map"),
            )
        ):
            self._ax_map = _AxisMapping({}, in_ndim=0)
        # required by ravel() and view() used in numpy testing.  Also for zeros_like...
        elif all(
            (
                isinstance(obj, AxesArray),
                hasattr(obj, "_ax_map"),
                not hasattr(self, "_ax_map"),
                self.shape == obj.shape,
            )
        ):
            self._ax_map = _AxisMapping(obj.axes, obj.ndim)
        # Using a poorly-initialized AxesArray
        # Occurs in MaskedArray.ravel, used in some plotting.  MaskedArray views
        # of AxesArray lose the axes attributes, and then the _ax_map attributes.
        # See numpy.ma.core:asanyarray
        elif all(
            (
                isinstance(obj, AxesArray),
                not hasattr(obj, "_ax_map"),
            )
        ):
            self._ax_map = _AxisMapping({"ax_unk": 0}, in_ndim=1)
        # maybe add errors for incompatible views?

    def __array_ufunc__(
        self, ufunc, method, *inputs, out=None, **kwargs
    ):  # this method is called whenever you use a ufunc
        args = []
        for input_ in inputs:
            if isinstance(input_, AxesArray):
                args.append(input_.view(np.ndarray))
            else:
                args.append(input_)

        outputs = out
        if outputs:
            out_args = []
            for output in outputs:
                if isinstance(output, AxesArray):
                    out_args.append(output.view(np.ndarray))
                else:
                    out_args.append(output)
            kwargs["out"] = tuple(out_args)
        else:
            outputs = (None,) * ufunc.nout
        results = super().__array_ufunc__(ufunc, method, *args, **kwargs)
        if results is NotImplemented:
            return NotImplemented
        if method == "at":
            return
        if ufunc.nout == 1:
            results = (results,)
        if method == "reduce" and (
            "keepdims" not in kwargs.keys() or kwargs["keepdims"] is False
        ):
            axes = None
            if kwargs["axis"] is not None:
                axes = self._ax_map.remove_axis(axis=kwargs["axis"])
        else:
            axes = self.axes
        final_results = []
        for result, output in zip(results, outputs):
            if output is not None:
                final_results.append(output)
            elif axes is None:
                final_results.append(result)
            else:
                final_results.append(AxesArray(np.asarray(result), axes))
        results = tuple(final_results)
        return results[0] if len(results) == 1 else results

    def __array_function__(self, func, types, args, kwargs):
        if func not in HANDLED_FUNCTIONS:
            return super(AxesArray, self).__array_function__(func, types, args, kwargs)
        if not all(issubclass(t, AxesArray) for t in types):
            return NotImplemented
        return HANDLED_FUNCTIONS[func](*args, **kwargs)


def _implements(numpy_function):
    """Register an __array_function__ implementation for AxesArray objects."""

    def decorator(func):
        HANDLED_FUNCTIONS[numpy_function] = func
        return func

    return decorator


@_implements(np.ravel)
def ravel(a, order="C"):
    out = np.ravel(np.asarray(a), order=order)
    is_1d_already = len(a.shape) == 1
    if is_1d_already:
        return AxesArray(out, a.axes)
    else:
        return AxesArray(out, {"ax_unk": 0})


@_implements(np.ix_)
def ix_(*args: AxesArray):
    calc = np.ix_(*(np.asarray(arg) for arg in args))
    ax_names = [list(arr.axes)[0] for arr in args]
    axes = fwd_from_names(ax_names)
    return tuple(AxesArray(arr, axes) for arr in calc)


@_implements(np.concatenate)
def concatenate(arrays, axis=0, out=None, dtype=None, casting="same_kind"):
    parents = [np.asarray(obj) for obj in arrays]
    ax_list = [obj.axes for obj in arrays if isinstance(obj, AxesArray)]
    for ax1, ax2 in zip(ax_list[:-1], ax_list[1:]):
        if ax1 != ax2:
            raise ValueError("Concatenating >1 AxesArray with incompatible axes")
    result = np.concatenate(parents, axis, out=out, dtype=dtype, casting=casting)
    if isinstance(out, AxesArray):
        out._ax_map = _AxisMapping(ax_list[0], in_ndim=result.ndim)
    return AxesArray(result, axes=ax_list[0])


@_implements(np.reshape)
def reshape(a: AxesArray, newshape: int | tuple[int], order="C"):
    """Gives a new shape to an array without changing its data.

    Args:
        a: Array to be reshaped
        newshape: int or tuple of ints
            The new shape should be compatible with the original shape.  In
            addition, the axis labels must make sense when the data is
            translated to a new shape.  Currently, the only use case supported
            is to flatten an outer product of two or more axes with the same
            label and size.
        order: Must be "C"
    """
    if order != "C":
        raise ValueError("AxesArray only supports reshaping in 'C' order currently.")
    out = np.reshape(np.asarray(a), newshape, order)  # handle any regular errors

    new_axes = {}
    if isinstance(newshape, int):
        newshape = [newshape]
    newshape = list(newshape)
    explicit_new_size = np.multiply.reduce(np.array(newshape))
    if explicit_new_size < 0:
        replace_ind = newshape.index(-1)
        newshape[replace_ind] = a.size // (-1 * explicit_new_size)

    curr_base = 0
    for curr_new in range(len(newshape)):
        if curr_base >= a.ndim:
            raise ValueError(
                "Cannot reshape an AxesArray this way.  Adding a length-1 axis at"
                f" dimension {curr_new} not understood."
            )
        base_name = a._ax_map.reverse_map[curr_base]
        if a.shape[curr_base] == newshape[curr_new]:
            compat_dict_append(new_axes, base_name, curr_new)
            curr_base += 1
        elif newshape[curr_new] == 1:
            raise ValueError(
                f"Cannot reshape an AxesArray this way.  Inserting a new axis at"
                f" dimension {curr_new} of new shape is not supported"
            )
        else:  # outer product
            remaining = newshape[curr_new]
            while remaining > 1:
                if a._ax_map.reverse_map[curr_base] != base_name:
                    raise ValueError(
                        "Cannot reshape an AxesArray this way.  It would combine"
                        f" {base_name} with {a._ax_map.reverse_map[curr_base]}"
                    )
                remaining, error = divmod(remaining, a.shape[curr_base])
                if error:
                    raise ValueError(
                        f"Cannot reshape an AxesArray this way.  Array dimension"
                        f" {curr_base} has size {a.shape[curr_base]}, must divide into"
                        f" newshape dimension {curr_new} with size"
                        f" {newshape[curr_new]}."
                    )
                curr_base += 1

            compat_dict_append(new_axes, base_name, curr_new)

    return AxesArray(out, axes=new_axes)


@_implements(np.transpose)
def transpose(a: AxesArray, axes: Optional[Union[Tuple[int], List[int]]] = None):
    """Returns an array with axes transposed.

    Args:
        a: input array
        axes: As the numpy function
    """
    out = np.transpose(np.asarray(a), axes)
    if axes is None:
        axes = range(a.ndim)[::-1]
    new_axes = {}
    old_reverse = a._ax_map.reverse_map
    for new_ind, old_ind in enumerate(axes):
        compat_dict_append(new_axes, old_reverse[old_ind], new_ind)

    return AxesArray(out, new_axes)


@_implements(np.einsum)
def einsum(
    subscripts: str, *operands: AxesArray, out: Optional[NDArray] = None, **kwargs
) -> AxesArray:
    calc = np.einsum(
        subscripts, *(np.asarray(arr) for arr in operands), out=out, **kwargs
    )
    try:
        # explicit mode
        lscripts, rscript = subscripts.split("->")
    except ValueError:
        # implicit mode
        lscripts = subscripts
        rscript = "".join(
            sorted(c for c in set(subscripts) if subscripts.count(c) == 1 and c != ",")
        )
    # 0-dimensional case, may just be better to check type of "calc":
    if rscript == "":
        return calc

    # assemble output reverse map
    allscript_names = _label_einsum_scripts(lscripts, operands)
    out_names = []

    for char in rscript.replace("...", "."):
        if char == ".":
            for script_names in allscript_names:
                out_names += script_names.get("...", [])
        else:
            ax_names = []
            for script_names in allscript_names:
                ax_names += script_names.get(char, [])
            ax_name = "ax_" + _join_unique_names(ax_names)
            out_names.append(ax_name)

    out_axes = fwd_from_names(out_names)
    if isinstance(out, AxesArray):
        out._ax_map = _AxisMapping(out_axes, calc.ndim)
    return AxesArray(calc, axes=out_axes)


def _join_unique_names(l_of_s: List[str]) -> str:
    ordered_uniques = dict.fromkeys(l_of_s).keys()
    return "_".join(
        ax_name[3:] if ax_name[:3] == "ax_" else ax_name for ax_name in ordered_uniques
    )


def _label_einsum_scripts(
    lscripts: List[str], operands: tuple[AxesArray]
) -> List[dict[str, str]]:
    """Create a list of what axis name each script refers to in its operand."""
    allscript_names: List[Dict[str, List[str]]] = []
    for lscr, op in zip(lscripts.split(","), operands):
        script_names: Dict[str, List[str]] = {}
        allscript_names.append(script_names)
        # handle script ellipses
        try:
            ell_ind = lscr.index("...")
            ell_width = op.ndim - (len(lscr) - 3)
            ell_expand = range(ell_ind, ell_ind + ell_width)
            ell_names = [op._ax_map.reverse_map[ax_ind] for ax_ind in ell_expand]
            script_names["..."] = ell_names
        except ValueError:
            ell_ind = len(lscr)
            ell_width = 0
        # handle script non-ellipsis chars
        shift = 0
        for ax_ind, char in enumerate(lscr):
            if char == ".":
                shift += 1
                continue
            if ax_ind < ell_ind:
                scr_name = op._ax_map.reverse_map[ax_ind]
            else:
                scr_name = op._ax_map.reverse_map[ax_ind - 3 + ell_width]
            compat_dict_append(script_names, char, [scr_name])
    return allscript_names


@_implements(np.linalg.solve)
def linalg_solve(a: AxesArray, b: AxesArray) -> AxesArray:
    result = np.linalg.solve(np.asarray(a), np.asarray(b))
    a_rev = a._ax_map.reverse_map
    a_names = [a_rev[k] for k in sorted(a_rev)]
    b_rev = b._ax_map.reverse_map
    b_names = [b_rev[k] for k in sorted(b_rev)]
    a_bcast_names = a_names[:-1]
    a_broadcast_ndim = a.ndim - 1
    if b.ndim == 1:
        b_broadcast_ndim = 1
        b_stack_axis_k = []
        b_bcast_names = b_names
        if b_names[0] != a_names[-2]:
            raise ValueError("Mismatch in operand axis names when aligning A and b")
    else:
        b_broadcast_ndim = b.ndim - 1
        b_stack_axis_k = [b_names[-1]]
        b_bcast_names = b_names[:-1]
    if a_bcast_names[-b_broadcast_ndim:] != b_bcast_names[-a_broadcast_ndim:]:
        raise ValueError("Mismatch in operand axis names when aligning A and b")
    bcast_names = (
        a_bcast_names if len(a_bcast_names) > len(b_bcast_names) else b_bcast_names
    )
    all_names = bcast_names[:-1] + [a_names[-1]] + b_stack_axis_k
    axes = fwd_from_names(all_names)
    return AxesArray(result, axes)


@_implements(np.tensordot)
def tensordot(
    a: AxesArray, b: AxesArray, axes: Union[int, Sequence[Sequence[int]]] = 2
) -> AxesArray:
    sub = _tensordot_to_einsum(a.ndim, b.ndim, axes)
    return einsum(sub, a, b)


def _tensordot_to_einsum(
    a_ndim: int, b_ndim: int, axes: Union[int, Sequence[Sequence[int]]]
) -> str:
    lc_ord = range(97, 123)
    sub_a = "".join([chr(code) for code in lc_ord[:a_ndim]])
    if isinstance(axes, int):
        axes = [range(-axes, 0), range(0, axes)]
    sub_b_li = [chr(code) for code in lc_ord[a_ndim : a_ndim + b_ndim]]
    if np.array(axes).max() > 26:
        raise ValueError("Too many axes")
    for a_ind, b_ind in zip(*axes):
        sub_b_li[b_ind] = sub_a[a_ind]
    sub_b = "".join(sub_b_li)
    sub = f"{sub_a},{sub_b}"  # noqa: E231
    return sub


def _standardize_indexer(
    arr: np.ndarray, key: Indexer | Sequence[Indexer]
) -> tuple[Sequence[StandardIndexer], tuple[KeyIndex, ...]]:
    """Convert any legal numpy indexer to a "standard" form.

    Standard form involves creating an equivalent indexer that is a tuple with
    one element per index of the original axis.  All advanced indexer elements
    are converted to numpy arrays, and boolean arrays are converted to
    integer arrays with obj.nonzero().

    Returns:
        A tuple of the normalized indexer as well as the indexes of
        advanced indexers
    """
    if isinstance(key, tuple):
        key = list(key)
    else:
        key = [key]

    if not any(ax_key is Ellipsis for ax_key in key):
        key = [*key, Ellipsis]

    new_key: List[Indexer] = []
    for ax_key in key:
        if not isinstance(ax_key, get_args(BasicIndexer)):
            ax_key = np.array(ax_key)
            if ax_key.dtype == np.dtype(np.bool_):
                new_key += ax_key.nonzero()
                continue
        new_key.append(ax_key)

    new_key = _expand_indexer_ellipsis(new_key, arr.ndim)
    # Can't identify position of advanced indexers before expanding ellipses
    adv_inds: List[KeyIndex] = []
    for key_ind, ax_key in enumerate(new_key):
        if isinstance(ax_key, np.ndarray):
            adv_inds.append(KeyIndex(key_ind))

    return new_key, tuple(adv_inds)


def _expand_indexer_ellipsis(key: List[Indexer], ndim: int) -> List[Indexer]:
    """Replace ellipsis in indexers with the appropriate amount of slice(None)"""
    # [...].index errors if list contains numpy array
    ellind = [ind for ind, val in enumerate(key) if val is ...][0]
    n_new_dims = sum(ax_key is None or isinstance(ax_key, str) for ax_key in key)
    n_ellipsis_dims = ndim - (len(key) - n_new_dims - 1)
    new_key = key[:ellind] + key[ellind + 1 :]
    new_key = new_key[:ellind] + (n_ellipsis_dims * [slice(None)]) + new_key[ellind:]
    return new_key


def _determine_adv_broadcasting(
    key: Sequence[StandardIndexer], adv_inds: Sequence[OldIndex]
) -> tuple[int, Optional[KeyIndex]]:
    """Calculate the shape and location for the result of advanced indexing."""
    adjacent = all(i + 1 == j for i, j in zip(adv_inds[:-1], adv_inds[1:]))
    adv_indexers = [np.array(key[i]) for i in adv_inds]
    bcast_nd = np.broadcast(*adv_indexers).nd
    bcast_start_axis = 0 if not adjacent else min(adv_inds) if adv_inds else None
    return bcast_nd, KeyIndex(bcast_start_axis)


def _rename_broadcast_axes(
    new_axes: List[tuple[int, None | str | Literal[_Sentinels.ADV_NAME]]],
    adv_names: List[str],
) -> List[tuple[int, str]]:
    """Normalize sentinel and NoneType names"""

    def _calc_bcast_name(*names: str) -> str:
        if not names:
            return ""
        if all(a == b for a, b in zip(names[1:], names[:-1])):
            return names[0]
        names = [name[3:] for name in dict.fromkeys(names)]  # ordered deduplication
        return "ax_" + "_".join(names)

    bcast_name = _calc_bcast_name(*adv_names)
    renamed_axes = []
    for ax_ind, ax_name in new_axes:
        if ax_name is None:
            renamed_axes.append((ax_ind, "ax_unk"))
        elif ax_name is _Sentinels.ADV_NAME:
            renamed_axes.append((ax_ind, bcast_name))
        else:
            renamed_axes.append((ax_ind, "ax_" + ax_name))
    return renamed_axes


def _replace_adv_indexers(
    key: Sequence[StandardIndexer],
    adv_inds: List[int],
    bcast_start_ax: int,
    bcast_nd: int,
) -> tuple[
    Union[None, str, int, Literal[_Sentinels.ADV_NAME], Literal[_Sentinels.ADV_REMOVE]],
    ...,
]:
    for adv_ind in adv_inds:
        key[adv_ind] = _Sentinels.ADV_REMOVE
    key = key[:bcast_start_ax] + bcast_nd * [_Sentinels.ADV_NAME] + key[bcast_start_ax:]
    return key


def _apply_indexing(
    key: tuple[StandardIndexer], reverse_map: Dict[int, str]
) -> tuple[
    List[int], List[tuple[int, None | str | Literal[_Sentinels.ADV_NAME]]], List[str]
]:
    """Determine where axes should be removed and added

    Only considers the basic indexers in key.  Numpy arrays are treated as
    slices, in that they don't affect the final dimensions of the output
    """
    remove_axes = []
    new_axes = []
    adv_names = []
    deleted_to_left = 0
    added_to_left = 0
    for key_ind, indexer in enumerate(key):
        if isinstance(indexer, int) or indexer is _Sentinels.ADV_REMOVE:
            orig_arr_axis = key_ind - added_to_left
            if indexer is _Sentinels.ADV_REMOVE:
                adv_names.append(reverse_map[orig_arr_axis])
            remove_axes.append(orig_arr_axis)
            deleted_to_left += 1
        elif (
            indexer is None
            or indexer is _Sentinels.ADV_NAME
            or isinstance(indexer, str)
        ):
            new_arr_axis = key_ind - deleted_to_left
            new_axes.append((new_arr_axis, indexer))
            added_to_left += 1
    return remove_axes, new_axes, adv_names


def comprehend_axes(x):
    axes = {}
    axes["ax_coord"] = len(x.shape) - 1
    axes["ax_time"] = len(x.shape) - 2
    if x.ndim > 2:
        axes["ax_spatial"] = list(range(len(x.shape) - 2))
    return axes


class SampleConcatter(TransformerMixin):
    def __init__(self):
        pass

    def fit(self, x_list, y_list=None):
        return self

    def __sklearn_is_fitted__(self):
        return True

    def transform(self, x_list):
        return concat_sample_axis(x_list)


def concat_sample_axis(x_list: List[AxesArray]):
    """Concatenate all trajectories and axes used to create samples."""
    new_arrs = []
    for x in x_list:
        sample_ax_names = ("ax_spatial", "ax_time", "ax_sample")
        sample_ax_inds = []
        for name in sample_ax_names:
            ax_inds = getattr(x, name, [])
            if isinstance(ax_inds, int):
                ax_inds = [ax_inds]
            sample_ax_inds += ax_inds
        new_axes = {"ax_sample": 0, "ax_coord": 1}
        n_samples = np.prod([x.shape[ax] for ax in sample_ax_inds])
        arr = AxesArray(x.reshape((n_samples, x.shape[x.ax_coord])), new_axes)
        new_arrs.append(arr)
    return np.concatenate(new_arrs, axis=new_arrs[0].ax_sample)


def wrap_axes(axes: dict, obj):
    """Add axes to object (usually, a sparse matrix)"""

    for key in ["ax_spatial", "ax_time", "ax_sample", "ax_coord"]:
        try:
            obj.__setattr__(key, axes[key])
        except KeyError:
            pass
    return obj


def compat_dict_append(
    compat_dict: CompatDict[T],
    key: str,
    item_or_list: ItemOrList[T],
) -> None:
    """Add an element or list of elements to a dictionary, preserving old values"""
    try:
        prev_val = compat_dict[key]
    except KeyError:
        compat_dict[key] = item_or_list
        return
    if not isinstance(item_or_list, list):
        item_or_list = [item_or_list]
    if not isinstance(prev_val, list):
        prev_val = [prev_val]
    compat_dict[key] = prev_val + item_or_list


def fwd_from_names(names: List[str]) -> CompatDict[int]:
    """Create mapping of name: axis or name: [ax_1, ax_2, ...]"""
    fwd_map: Dict[str, Sequence[int]] = {}
    for ax_ind, name in enumerate(names):
        compat_dict_append(fwd_map, name, [ax_ind])
    return fwd_map



import numpy as np

import pysindy as ps


class InferenceBM:
    """
    An example benchmark that times the performance of various kinds
    of iterating over dictionaries in Python.
    """

    def setup(self):
        t = np.linspace(0, 1, 100)
        x = 3 * np.exp(-2 * t)
        y = 0.5 * np.exp(t)
        X = np.stack((x, y), axis=-1)  # First column is x, second is y

        self.model = ps.SINDy()
        self.model.fit(X, t=t, feature_names=["x", "y"])
        rng = np.random.default_rng(42)
        self.inputs = rng.random((1000, 2))

    def time_a_predict(self):
        self.model.predict(self.inputs[:1])

    def peakmem_a_predict(self):
        self.model.predict(self.inputs[:1])

    def time_batch_predict(self):
        self.model.predict(self.inputs)

    def peakmem_batch_predict(self):
        self.model.predict(self.inputs)
